from  ...mechanics import *
from  ...mechanics import ReportComponent as BaseReportComponent


from . import pl

import pandas as pd
import numpy as np
from sympy import lambdify

from sympy import *
from pandas import *
from sympy.physics.mechanics import dynamicsymbols

from .....solvers.linear import *
from .....dynamics import *

months_list = ['January', 'February', 'March','April','May','June','July','August','September','October','November','December']

from  ...mechanics import display
import datetime


miesiace_list = ['styczeń', 'luty', 'marzec','kwiecień','maj','czerwiec','lipiec','sierpień','wrzesień','październik','listopad','grudzień']

average_temp_list = [-1.9,-0.8,3.2,9.3,14.6,18,20.1,19.5,14.7,9.3,4.8,0.5]

Eg_daily_list_Wh_per_meter2 =[600,1000,3000,3800,4800,5400,5300,4900,3300,1700,700,500]

Eg_daily_list_kWh_per_meter2 = [0.6,1,3,3.8,4.8,5.3,4.9,3.3,1.7,0.7,0.5]

day_length_in_months_hours = [8.3,10.0,11.8,13.9,15.7,16.7,16.3,14.7,12.7,10.7,8.8,7.8]

data_atmospheric_conditions = {'Day length in a month [h]':day_length_in_months_hours,'Daily energy intensity [${kWh/m^2}$]':Eg_daily_list_Wh_per_meter2,'Average temperature [$^{\circ}$C]':average_temp_list}

df = pd.DataFrame(index = months_list,data = data_atmospheric_conditions)

class ReportComponent(BaseReportComponent):



    @property
    def reported_object(self):

        from .....solvers.linear import ODESystem
        
        if isinstance(self._reported_object, ODESystem):
            return self._reported_object

        else:
            return self._reported_object
            
    @reported_object.setter
    def reported_object(self, obj):
        self._reported_object=obj

    @property
    def _system(self):
        print('Kod do poprawienia #################### bo stosujesz starą zmienną self._system')
        print('zamień se self._system na self.reported_object')
        
        return self.reported_object


#pandas guide
data_code=(
'''
months_list = ['January', 'February', 'March','April','May','June','July','August','September','October','November','December']
average_temp_list = [-1.9,-0.8,3.2,9.3,14.6,18,20.1,19.5,14.7,9.3,4.8,0.5]
day_length_in_months_hours = [8.3,10.0,11.8,13.9,15.7,16.7,16.3,14.7,12.7,10.7,8.8,7.8]
Eg_daily_list_Wh_per_meter2 =[600,1000,3000,3800,4800,5400,5300,4900,3300,1700,700,500]
Eg_daily_list_kWh_per_meter2 = [0.6,1,3,3.8,4.8,5.3,4.9,3.3,1.7,0.7,0.5]
''')

data_dict_code=(
'''
data_atmospheric_conditions = {'Day length in a month [h]':day_length_in_months_hours,'Daily energy intensity [${kWh/m^2}$]':Eg_daily_list_Wh_per_meter2,'Average temperature [$^{\circ}$C]':average_temp_list}
''')
output_code=(
'''
import pandas as pd

df = pd.DataFrame(index = months_list,data = data_atmospheric_conditions)
df
''')

class PandasTableGenerationComponent(ReportComponent):
    
    title="Generating a table"


    def append_elements(self):
        
        #system = self.reported_object # it's useless in the case of permanent content - it's commented for future usage


        display(ReportText('Creating a list of data:'))

        display(GuideCode(data_code))


        display(ReportText('Creating dictionary:'))

        display(GuideCode(data_dict_code))


        display(ReportText('Calling the table:'))
        display(GuideCode(output_code))
        display(df)
        


class PandasMethodsComponent(ReportComponent):
    
    title="Pandas library methods"


    def append_elements(self):
        
        #system = self.reported_object # it's useless in the case of permanent content - it's commented for future usage
        
        
        months_list = ['January', 'February', 'March','April','May','June','July','August','September','October','November','December']

        average_temp_list = [-1.9,-0.8,3.2,9.3,14.6,18,20.1,19.5,14.7,9.3,4.8,0.5]

        Eg_daily_list_Wh_per_meter2 =[600,1000,3000,3800,4800,5400,5300,4900,3300,1700,700,500]

        Eg_daily_list_kWh_per_meter2 = [0.6,1,3,3.8,4.8,5.3,4.9,3.3,1.7,0.7,0.5]

        day_length_in_months_hours = [8.3,10.0,11.8,13.9,15.7,16.7,16.3,14.7,12.7,10.7,8.8,7.8]

        data_atmospheric_conditions = {'Day length in a month [h]':day_length_in_months_hours,'Daily energy intensity [${kWh/m^2}$]':Eg_daily_list_Wh_per_meter2,'Average temperature [$^{\circ}$C]':average_temp_list}

        df = pd.DataFrame(index = months_list,data = data_atmospheric_conditions)

        display(ReportText('Iloc method:'))
        display(GuideCode('df.iloc[0:3]'))
        display(df.iloc[0:3])

        # display(Picture('./Image/iloc.jpg', caption = ""))

        display(ReportText('Loc method:'))
        display(GuideCode('df.loc["January"]'))
        display(df.loc["January"])
        # display(Picture('./Image/loc.jpg', caption = ""))

        display(ReportText('The set axis method works for both columns and rows. For rows:'))
        display(GuideCode("df.set_axis(['a','b','c','d','e','f','g','h','i','j','k','l'],axis = 'index')"))
        display(df.set_axis(['a','b','c','d','e','f','g','h','i','j','k','l'], axis = 'index'))

        # display(Picture('./Image/set_axis.jpg', caption = ""))

        display(ReportText('For columns:'))
        display(GuideCode("df.set_axis(['a','b','c'],axis = 'columns')"))
        display(df.set_axis(['a','b','c'],axis = 'columns'))
        # display(Picture('./Image/set_axis2.jpg', caption = ""))

        display(ReportText('Metoda rename - Zmienienie pojedynczej kolumny:'))
        display(GuideCode('''df.rename(columns = {"Day length in a month [h]":'AAA'}, index = {"January":'A'} )'''))
        display(df.rename(columns = {"Day length in a month [h]":'AAA'}, index = {"January":'A'} ))
        # display(Picture('./Image/rename.jpg', caption = ""))

        display(ReportText('Map method:'))
        display(GuideCode('df.map(lambda x:x+2)'))
#         display(df.map(lambda x:x+2)) - nowa wersja pandasa zmieniła nazwę "map" na "applymap" - Michał SZ
        display(df.map(lambda x:x+2))

        # display(ReportText('***Applymap mthod for colummn/row:***'))
        # display(Markdown('''
        # Remeber, that df[NAME] => creates series, a df[[NAME]] => creates new frame, which can be used with .applymap'''))
        # display(df[['Day length in a month']].applymap(lambda x:x+100))
        # pd_h = df[['Day length in a month']].applymap(lambda x:x+100)
        # list1 = pd_h['Day length in a month'].tolist()
        # df['Day length in a month'] = list1

        # display(df)

        # display(Picture('./Image/applymap.jpg', caption = ""))

        display(ReportText('Slicing:'))
        display(GuideCode('''df['Day length in a month [h]']'''))
        display(df['Day length in a month [h]'])
        # display(Picture('./Image/slice2.jpg', caption = ""))
        

        
class NumericalAnalysisSimulationComponent(pl.NumericalAnalysisSimulationComponent):
    
    title="Numerical analysis simulation Component"

class AnalyticalSimulationComponent(pl.AnalyticalSimulationComponent):
    
    title="Analytical simulation Component"
    

doc_gen_str =(
'''

doc_final = MechanicalCase('./output/nazwa_dokumentu',documentclass=NoEscape('article'),document_options=['a4paper','fleqn'],lmodern=False)
doc_final.packages.append(Package('natbib', options=['numbers']))
doc_final.packages.append(Package('booktabs'))
doc_final.packages.append(Package('float'))
doc_final.packages.append(Package('siunitx'))


doc_final.append(sekcja)
doc_final.append(sekcja1)
doc_final.append(sekcja2)
doc_final.append(sekcja3)
doc_final.append(sekcja4)
doc_final.append(sekcja5)

doc_final.generate_pdf()

''')

class DocumentGenerationComponent(ReportComponent):

    #title="Generowanie dokumentu"
    title="Document generation"


    @property
    def reported_object(self):


        default_data = {'classname':'ReportingModuleIntroComponent',
                       'module':'guide.en.py',
                       'field':'guide or report',
                       'target':'`ODESystem` class',
                       'issue_no':359,
                       }


        if isinstance(self._reported_object, dict):
            return {**default_data,**self._reported_object}

        elif isinstance(self._reported_object, str):
            return {**default_data,'classname':self._reported_object}

        elif self._reported_object is None:
            return default_data

        elif self._reported_object is None or not isinstance(self._reported_object, dict):
            return default_data
        
        else:
            return self._reported_object


    @reported_object.setter
    def reported_object(self, obj):
        self._reported_object=obj


    def append_elements(self):
        #variables provided by `reported_object` arg
        classname = self.reported_object['classname']
        class_module = self.reported_object['module']
        class_field = self.reported_object['field']
        target = self.reported_object['target']
        
       
        #implement reporting activieties here
        
        #display(ReportText('Ostatni krok to zaapendowanie sekcji i utworzenie dokumentu pdf :'))
        display(ReportText('The last step is to append a section and create a pdf document :'))
        display(GuideCode(f'{doc_gen_str}'))


        
TablesCreationComponent_str=(
'''
from dynpy.models.mechanics.tmac import SDOFWinchSystem
eoms= SDOFWinchSystem()._eoms[0]
eoms_eq_raw=Eq(eoms,0)
data=SDOFWinchSystem().get_random_parameters()
Ms0=Symbol('M_s0')
b=Symbol('b')
omg=Symbol('varphi')
ph1=Symbol('Phi')

winch = SDOFWinchSystem()

Ms=winch.M_s
t = winch.ivar

dane={
    Ms0:4700,
    b:1500,
    SDOFWinchSystem().m0:10}



winch_up=SDOFWinchSystem().subs(Ms,Ms0-b*SDOFWinchSystem().phi.diff()).subs(data).subs(dane)
nowe_ode=winch_up._ode_system
solution=nowe_ode.solution
steady_sol=nowe_ode.steady_solution[0]
t_span = np.linspace(0,100,200)
phi=winch.q[0]



units_dict = {phi : ureg.radian,phi.diff(t):ureg.radian,t:ureg.second }
LatexDataFrame.set_default_units(units_dict)

general_sol_matrix=nowe_ode.solution.with_ics([0,0])
table_eq=general_sol_matrix.n().numerized().compute_solution(t_span)
table=table_eq[[phi.diff(t)]]
table_for_report = table[0::15].to_latex_dataframe()

display(Markdown(' Do tablicy `table_eq` zapisujemy równanie ogólne dla zmiennej czasowej `t_span`   '))
display(Markdown(
f"""
    
    table_eq=general_sol_matrix.n().numerized().compute_solution(t_span)

"""))
display(Markdown(f' Do tablicy table przypisujemy wartości ${latex(omg)}$ powstałe w wyniku zróżniczkowania ${latex(ph1)}$ w tablicy `table_eq` '))
display(Markdown(
f"""
    table=table_eq[[phi.diff(t)]]
   
"""))
display(ReportText(' W następnym kroku wybieramy zakres wartości który chcemy umieścić w tabelce przy pomocy naiwasów kwadratowych. Nastepnie przekształcamy tabelę na zgodną z LateX przy pomocy metody: to\_latex\_dataframe()  '))
display(Markdown(
f"""

    table_for_report = table[0::15].to_latex_dataframe()

"""))
display(ReportText(' Naszą tabelkę opisujemy w tekście, pożądane jest użycie AutoMarkerów które zostały opisane w poniższej sekcji tego dokument '))
display(Markdown(
f"""

     display(Markdown(f'Wyniki symulacji zostały zestawione w tabeli {AutoMarker(table_for_report)}'))

"""))
display(ReportText(' Aby wyświetlic tabelkę korzystamy z  klasy display. Metoda .reported umożliwia nam nadanie tabelce podpisu. '))
display(Markdown(
f"""

    display(table_for_report.reported(caption='Tabela'))

"""))
    
''')

class TablesCreationComponent(DocumentGenerationComponent):

    title="Tworzenie tabelek"


    def append_elements(self):
        #variables provided by `reported_object` arg
        classname = self.reported_object['classname']
        class_module = self.reported_object['module']
        class_field = self.reported_object['field']
        target = self.reported_object['target']
        
       
        #implement reporting activieties here
        
        display(ReportText('Aby utworzyć tabelkę najpierw musimy posiadać dane z obliczeń na interesującym nas systemie dynamicznym:'))
        display(GuideCode(f'{TablesCreationComponent_str}'))

SympyFormulaComponent_str='''
from sympy import *

M = Symbol('M')

t = Symbol('t')

m_e = Symbol('m_e')

e = Symbol('e')

z = Function('z')

varphi = Function('varphi')

Ek = M*Derivative(z(t), t)**2/2 + m_e*(e*sin(varphi(t))*Derivative(varphi(t), t) - Derivative(z(t), t))**2/2

display(Eq(Symbol('T'),Ek))
'''
class AlgebraicExpressionComponent(ReportComponent):

    title = "Algebraic expression example"

    def append_elements(self):
        display(ReportText('Example Algebraic expression in Sympy library:'))
        display(GuideCode(f'{SympyFormulaComponent_str}'))
        
        from sympy import Symbol, Function

        M = Symbol('M')

        t = Symbol('t')

        m_e = Symbol('m_e')

        e = Symbol('e')

        z = Function('z')

        varphi = Function('varphi')

        Ek = M*Derivative(z(t), t)**2/2 + m_e*(e*sin(varphi(t))*Derivative(varphi(t), t) - Derivative(z(t), t))**2/2
        
        display(ReportText('After executing above code you can display kinetic energy algebraic expression:'))
        
        display(Eq(Symbol('T'),Ek))


pandas_intro_call_str=("""
from dynpy.utilities.documents.guides import IntroToPandasGuide
from dynpy.utilities.components.guides.en import *
from dynpy.models.mechanics.pendulum import ForcedSpringMassSystem

IntroToPandasGuide(reported_object=ForcedSpringMassSystem());

""")


pandas_intro_str =('''Guide consist of bellow components which can be called separatelly:
*IntroToPandasUsageComponent
*PandasTableGenerationComponent
*PandasMethodsComponent
*BasicOperationsComponent
*DynamicSystemCallComponent
*SimulationsComponent
*DifferentSimulationsComponent
''')

class IntroToPandasUsageComponent(ReportComponent):

    title = "Introduction to basics of pandas library"

    def append_elements(self):

        from dynpy.models import mechanics


        display(ReportText('This guide concers basic usage of pandas library with dynamical systems simulations in `DynPy` library.'))
        display(ReportText('Basic call of it is as follows and runs default dynamic system which is `ForcedSpringMassSystem'))
        display(ObjectCode(pandas_intro_call_str))
        
        display(ReportText(pandas_intro_str))




from  ...mechanics import *
from  ...mechanics import ReportComponent as BaseReportComponent


from . import pl

import pandas as pd
import numpy as np
from sympy import lambdify

from sympy import *
from pandas import *
from sympy.physics.mechanics import dynamicsymbols

from .....solvers.linear import *
from .....dynamics import *

months_list = ['January', 'February', 'March','April','May','June','July','August','September','October','November','December']

from  ...mechanics import display
import datetime


miesiace_list = ['styczeń', 'luty', 'marzec','kwiecień','maj','czerwiec','lipiec','sierpień','wrzesień','październik','listopad','grudzień']

average_temp_list = [-1.9,-0.8,3.2,9.3,14.6,18,20.1,19.5,14.7,9.3,4.8,0.5]

Eg_daily_list_Wh_per_meter2 =[600,1000,3000,3800,4800,5400,5300,4900,3300,1700,700,500]

Eg_daily_list_kWh_per_meter2 = [0.6,1,3,3.8,4.8,5.3,4.9,3.3,1.7,0.7,0.5]

day_length_in_months_hours = [8.3,10.0,11.8,13.9,15.7,16.7,16.3,14.7,12.7,10.7,8.8,7.8]

data_atmospheric_conditions = {'Day length in a month [h]':day_length_in_months_hours,'Daily energy intensity [${kWh/m^2}$]':Eg_daily_list_Wh_per_meter2,'Average temperature [$^{\circ}$C]':average_temp_list}

df = pd.DataFrame(index = months_list,data = data_atmospheric_conditions)

class ReportComponent(BaseReportComponent):



    @property
    def reported_object(self):

        from .....solvers.linear import ODESystem
        
        if isinstance(self._reported_object, ODESystem):
            return self._reported_object

        else:
            return self._reported_object
            
    @reported_object.setter
    def reported_object(self, obj):
        self._reported_object=obj

    @property
    def _system(self):
        print('Kod do poprawienia #################### bo stosujesz starą zmienną self._system')
        print('zamień se self._system na self.reported_object')
        
        return self.reported_object


# 
#Zakladanie Jupytera        
# 
dynpy_imports_code=(
'''

from sympy import* 
from sympy.physics.mechanics import dynamicsymbols, init_vprinting
from sympy.abc import* 
init_vprinting()
from sympy import S
from dynpy.solvers.linear import ODESystem
import pandas as pd
from dynpy.models.mechanics import Pendulum
from dynpy.models import mechanics

from pylatex import Document, Section, Subsection, Itemize, Package,  HorizontalSpace, Description, Marker, Command
from pylatex.section import Paragraph, Chapter
from pylatex.utils import italic, NoEscape
from dynpy.utilities.adaptable import *
from dynpy.utilities.templates.document import BeamerTemplate, MechanicalCase, EngeneeringDrawingGuide,DevelopmentGuide
from dynpy.utilities.templates.tikz import TikzCaSCStandalone
from dynpy.utilities.report import ReportText, Markdown, Picture, SympyFormula, Frame, ObjectCode, Block, AlertBlock, ExampleBlock
from dynpy.utilities.report import (SystemDynamicsAnalyzer, DataPlot, AccelerationComparison, FFTComparison, ReportEntry,             SimulationalBlock, ReportText, SimulationFFT, DataStorage, Markdown, SummaryTable, SympyFormula, SymbolsDescription, DescriptionsRegistry,             ObjectCode,CurrentContainer)
from dynpy.models import mechanics
import inspect

''')

class JupyterSetUpComponent(ReportComponent):
    
    title="Zakładanie Jupytera"


    def append_elements(self):

        #system = self.reported_object # it's useless in the case of permanent content - it's commented for future usage
        display(ReportText('Posługując sie Jupyterem będziemy tworzyć cele i wykonywali w nich pewne zadania.  '))

        pic9 = Picture('./dynpy/utilities/components/guides/images/PoDodaniuCeli.png', width='9cm')

        display(pic9)

        display(ReportText('Aby dodac nową cele wystarczy kliknąć na "prostokąt" i pod nim pojawią się opcje takie jak Code czy Text'))


        pic10 = Picture('./dynpy/utilities/components/guides/images/Markdown.png', width='9cm')

        display(pic10)

        display(ReportText('Po wybraniu opcji Text będziemy mogli napisać opis zadania, które wykonujemy lub nadać sekcji nagłówek  '))

        pic11 = Picture('./dynpy/utilities/components/guides/images/Plotki-w-naglowk.png', width='9cm')


        display(ReportText('Aby dodać nagłowek wystarczy wpisac znak Hashtag/płotek, ważne jest to że każdy nagłówek będzie traktowany jako sekcja niższego stopnia (\# główna sekcja, \#\# podsekcja itd...)'))

        display(pic11)

        pic12 = Picture('./dynpy/utilities/components/guides/images//Nazywanie-naglowka.png', width='9cm')

        display(pic12)


        display(ReportText('Po wejsciu w zakładkę Edit ukażą nam się poniższe opcje, takie jak usuwanie cel.'))

        pic13 = Picture('./dynpy/utilities/components/guides/images/Del-inne-opcje.png', width='9cm')

        display(pic13)


        display(ReportText('Czasami gdy Jupyter się zawiesi można spróbować go zrestartować uzywając poniższego przycisku'))

        pic14 = Picture('./dynpy/utilities/components/guides/images/Restart.jpeg', width='9cm')

        display(pic14)

        pic15 = Picture('./dynpy/utilities/components/guides/images/Restart-kernel.png', width='9cm')

        display(pic15)

        display(ReportText('Aby mieć pewność że w naszym kodzie uwzględnione zostały wszystkie zaimportowane biblioteki warto uruchamiając jupytera użyć opcji Run all cells pokazanej na poniższych obrazkach'))

        pic16 = Picture('./dynpy/utilities/components/guides/images/RunAll.jpeg', width='9cm')

        display(pic16)

        pic17 = Picture('./dynpy/utilities/components/guides/images/Run_all.png', width='9cm')

        display(pic17)

        display(ReportText('Aby elementy z których korzystamy działały, należy zaimportować potrzebne biblioteki, takie jak na załączonym obrazku oraz w formie tekstu który można skopiowac'))

        pic18 = Picture('./dynpy/utilities/components/guides/images/ImportowanieBibliotek.png', width='9cm')

        # display(pic18)

        display(GuideCode(dynpy_imports_code))

        
#obsluga systemow dynamicznych


imports_code_str = '''
from sympy import *
import numpy as np
from pandas import *
from sympy.physics.mechanics import init_vprinting, dynamicsymbols
from pylatex import Document, Section, Subsection, Itemize, Package, HorizontalSpace, Description, Marker, Command 
from pylatex.section import Paragraph, Chapter 
from pylatex.utils import italic, NoEscape 
from dynpy.utilities.adaptable import *
from dynpy.utilities.templates.document import BeamerTemplate, MechanicalCase 
from dynpy.utilities.templates.tikz import TikzCaSCStandalone
from dynpy.utilities.report import (Markdown, Picture, SympyFormula, SymbolsDescription, DescriptionsRegistry, ObjectCode, CurrentContainer)
from dynpy.solvers.linear import ODESystem
init_vprinting()
from dynpy.models.mechanics import SDOFWinchSystem

'''



class DynamicSystemCallComponent(ReportComponent):
    
    title="Introduction"


    def append_elements(self):

        system = self.reported_object # it's useless in the case of permanent content - it's commented for future usage
        system_name = system.__class__.__name__
        
        display(ReportText('Required library/class imports to create the report:'))


        display(GuideCode(imports_code_str.replace('SDOFWinchSystem',system_name)))


        display(ReportText('A class that allows you to solve a specific problem (in this case, a problem related to the description of a crane):'))

        display(GuideCode("""from dynpy.models.mechanics import SDOFWinchSystem""".replace('SDOFWinchSystem',system_name)      ))
        
        
        display(GuideCode(f'system=dyn_sys={system_name}()'  ))

        display(ReportText('Path to the sought-after class on the CoCalc platform:'))
        #display(Picture('./dynpy/utilities/components/guides/images/sciezka_w.jpg'))

        display(ReportText((system.__class__.__module__)))
        
        display(ReportText('The way to call out the preview of the class - the so-called preview:'))

        display(GuideCode(f'''{system_name}()._as_picture()'''  ))
        display(system._as_picture())
        

    
class DynamicSystemCompletenessCheckComponent(pl.DynamicSystemCompletenessCheckComponent):
    
    title="Dynamic system completeness check component"

        
## TABELKA InterimSchedule 

spotkanie_1='08.10.24'
spotkanie_2='15.10.24'
spotkanie_3='22.10.24'
spotkanie_4='29.10.24'
spotkanie_5='05.11.24'
spotkanie_6='12.11.24'
spotkanie_7='19.11.24'
spotkanie_8='26.11.24'
spotkanie_9='03.12.24'
spotkanie_10='10.12.24'


zadanie_1='Ustalenie tematu pracy przejściowej i wstępnego toku postępowania'
zadanie_2='Wyprowadzenie równań opisujących zjawisko'
zadanie_3='Zaimplementowanie równań w środowisku obliczeniowym'
zadanie_4='Dodanie opisu oraz listy symboli występujących w równaniach'
zadanie_5='Przeprowadzenie badań/symulacji, uzyskanie danych pomiarowych)'
zadanie_6='Wykonanie tabeli, przetworzenie jej na wykresy, opis zachodzących zależności'
zadanie_7='Opracowanie wniosków, wyników odnośnie przeprowadzonych badań/symulacji'
zadanie_8='Formowanie wstępu  oraz podsumowania pracy '
zadanie_9='Poprawa błędów'
zadanie_10='Poprawa błędów, oddanie pracy'

wl='Praca zasadnicza'
po='Poprawki'
        
class InterimScheduleComponent(ReportComponent):
    
    title="Implementacja systemow dynamicznych"


    @property
    def reported_object(self):

        default_data = {'date':datetime.datetime(2024,7,16),
                       'title':'Title of interim project',
                       'timedelta':datetime.timedelta(days=7),
                       'issue_no':359,
                       }

    
        
        if isinstance(self._reported_object, dict):
            return {**default_data,**self._reported_object}

#         elif isinstance(self._reported_object, str):
#             return {**default_data,'date':self._reported_object}
        elif isinstance(self._reported_object, str):
            return {**default_data, 'date': datetime.datetime.strptime(self._reported_object, "%Y-%m-%d")}
                    
        elif self._reported_object is None:
            return default_data

        else:
#             return self._reported_object
            return {**default_data, 'date': self._reported_object}


    @reported_object.setter
    def reported_object(self, obj):
        self._reported_object=obj


    def append_elements(self):
        
        first_meeting_date = self.reported_object['date'] # it's useless in the case of permanent content - it's commented for future usage
        days_inc = self.reported_object['timedelta']


        #przekopiowac tu fragment kodu z guide pamietajac o indentach
        display(ReportText(f"Pracę rozpoczynamy {first_meeting_date}")) #itp
        
        tabela_git = {
        'Tydzień prac': [1,2,3,4,5,6,7,8,9,10],
        'Data': [first_meeting_date+datetime.timedelta(days=days_inc.days*no) for no in  range(10)],
        'Temat konsultacji': [zadanie_1, zadanie_2, zadanie_3, zadanie_4, zadanie_5, zadanie_6, zadanie_7, zadanie_8, zadanie_9, zadanie_10],
        'Charakter spotkania': [wl,wl,wl,wl,wl,wl,wl,wl,po,po]
            
        
}

        tabelka_pandas=pd.DataFrame(tabela_git)
        tabelka_pandas.index += 1
        tabelka_latex=LatexDataFrame(tabelka_pandas).reported(caption='Harmonogram spotkań')
        display((display(tabelka_latex)))

class InterimTemplateComponent(InterimScheduleComponent):
    
    title="Implementacja systemow dynamicznych"


    
    
    def append_elements(self):
        
        #system = self.reported_object # it's useless in the case of permanent content - it's commented for future usage


        #przekopiowac tu fragment kodu z guide pamietajac o indentach
        display(ReportText("Tu kod dla podstawowej formy dokumentu")) #itp
    
    
    
project_issue_title_str = """
Maintenance of `{system_name}` class which is dynamic system representation        
"""
        
project_issue_desc_str = """

The following problems have to be checked or fixed, in order to ensure the correctness of implemented system:

- [ ] checking if a class is executable,

- [ ] validation of figures for schemes and real examples,

- [ ] validation of reference parameters,

- [ ] validation of random parameters,

- [ ] validation of  description of  description of parameters,

- [ ] validation of units.
"""
        
class InterimIssuesComponent(InterimScheduleComponent):
    
    title="Wzór na tworzenie issues w systemie GitHub dla przejsciowe"

    def append_elements(self):
        
        system = self.reported_object # it's useless in the case of permanent content - it's commented for future usage
        system_name = system.__class__.__name__
        
        display(ReportText("Przykładowy tekst, który pozwoli na sprawne przygotowywanie issue ma następującą formę:"))

        display(ReportText("# Title: "))
        display(ObjectCode(project_issue_title_str.format(system_name=system_name)))

        display(ReportText("# Description: "))
        display(ObjectCode(project_issue_desc_str))

        
        
######### GUIDE REPORTING COMPONENTS
######### GUIDE REPORTING COMPONENTS
######### GUIDE REPORTING COMPONENTS

doc_gen_str =(
'''

doc_final = MechanicalCase('./output/nazwa_dokumentu',documentclass=NoEscape('article'),document_options=['a4paper','fleqn'],lmodern=False)
doc_final.packages.append(Package('natbib', options=['numbers']))
doc_final.packages.append(Package('booktabs'))
doc_final.packages.append(Package('float'))
doc_final.packages.append(Package('siunitx'))


doc_final.append(sekcja)
doc_final.append(sekcja1)
doc_final.append(sekcja2)
doc_final.append(sekcja3)
doc_final.append(sekcja4)
doc_final.append(sekcja5)

doc_final.generate_pdf()

''')

class DocumentGenerationComponent(ReportComponent):

    #title="Generowanie dokumentu"
    title="Document generation"


    @property
    def reported_object(self):


        default_data = {'classname':'ReportingModuleIntroComponent',
                       'module':'guide.en.py',
                       'field':'guide or report',
                       'target':'`ODESystem` class',
                       'issue_no':359,
                       }


        if isinstance(self._reported_object, dict):
            return {**default_data,**self._reported_object}

        elif isinstance(self._reported_object, str):
            return {**default_data,'classname':self._reported_object}

        elif self._reported_object is None:
            return default_data

        elif self._reported_object is None or not isinstance(self._reported_object, dict):
            return default_data
        
        else:
            return self._reported_object


    @reported_object.setter
    def reported_object(self, obj):
        self._reported_object=obj


    def append_elements(self):
        #variables provided by `reported_object` arg
        classname = self.reported_object['classname']
        class_module = self.reported_object['module']
        class_field = self.reported_object['field']
        target = self.reported_object['target']
        
       
        #implement reporting activieties here
        
        #display(ReportText('Ostatni krok to zaapendowanie sekcji i utworzenie dokumentu pdf :'))
        display(ReportText('The last step is to append a section and create a pdf document :'))
        display(GuideCode(f'{doc_gen_str}'))


        
from dynpy.utilities.components.mech.en import KineticEnergyComponent, KineticEnergySymPyCodeComponent



kod_2='''
from dynpy.utilities.components.guides.en import ReportComponent
    from dynpy.utilities.report import ReportText, Markdown, Picture, SympyFormula, Frame, ObjectCode, Block, AlertBlock, ExampleBlock, GuideCode, LatexDataFrame
    from dynpy.utilities.components.guides.en import ReportCompImplementationComponent,ReportCompImplementationIssueComponent
    from sympy import *
    from pint import UnitRegistry
    import pandas as pd
'''
class LibrariesImportComponent(DocumentGenerationComponent):

    #title="Klasa LibrariesImportComponent i jej zastosowanie"
    title="LibrariesImportComponent class and its use"

    def append_elements(self):
        classname = self.reported_object['classname']
        class_module = self.reported_object['module']
        class_field = self.reported_object['field']
        target = self.reported_object['target']
        #display(ReportText('Przed pracą w jupiterze należy zaimportować biblioteki'))
        display(ReportText('Before working in jupiter, import libraries'))
        display(ObjectCode(kod_2))
        #display(ReportText('Dzięki temu możemy korzystać z przygotowanych wcześniej klas i modułów'))
        display(ReportText('This allows us to use previously prepared classes and modules'))

        
        
class ModellingInPythonGuidelinesComponent(ReportComponent):

    #title="Klasa LibrariesImportComponent i jej zastosowanie"
    title="Schedule of course on Advanced Modelling In Python"

    def append_elements(self):
        #classname = self.reported_object['classname']
        #class_module = self.reported_object['module']
        #class_field = self.reported_object['field']
        #target = self.reported_object['target']
        
        #display(ReportText('Przed pracą w jupiterze należy zaimportować biblioteki'))
        display(ReportText('Crucial information about the course on Advanced Modelling In Python'))
        display(ObjectCode(kod_2))
        #display(ReportText('Dzięki temu możemy korzystać z przygotowanych wcześniej klas i modułów'))
        display(ReportText('exemplary entry - fill as needed'))
        
        
class ModuleStructureComponent(ReportComponent):

    title="Introduction to ModuleStructure class"
    def append_elements(self):
        display(ReportText('ModuleStructure is a class that can be used to report dynsys class structure. There are few methods implemented. First method get_classes will return list of classes and modules. It can be called using bellow code:'))
        display(ObjectCode('''import dynpy
from dynpy.utilities.creators import ModuleStructure

ModuleStructure('dynpy.utilities.components').get_classes()
        '''))
        display(ReportText('You can also display this sstructure as tree using bellow method:'))
        display(ObjectCode('''import dynpy
from dynpy.utilities.creators import ModuleStructure

display(ObjectCode(ModuleStructure(dynpy.utilities.components).get_module_tree()))
        '''))
        display(ReportText('This class can also be used for updating init files in submodules. Bellow call will outout ready to use text for init file:'))
        display(ObjectCode('''from dynpy.utilities.creators import ModuleStructure

ModuleStructure.get_init_file_content()
        '''))
        display(ReportText('Another usefull feature might be geting import command for any class used in dynpy'))
        display(ObjectCode('''from dynpy.utilities.creators import ModuleStructure

ModuleStructure.get_import('ObjectCode')
        '''))
        
class ModellingInPythonScheduleComponent(ReportComponent):

    #title="Klasa LibrariesImportComponent i jej zastosowanie"
    title="Schedule of course on Advanced Modelling In Python"

    def append_elements(self):
        #classname = self.reported_object['classname']
        #class_module = self.reported_object['module']
        #class_field = self.reported_object['field']
        #target = self.reported_object['target']
        
        #display(ReportText('Przed pracą w jupiterze należy zaimportować biblioteki'))
        display(ReportText('Crucial information about the course schedule....'))
        display(ObjectCode(kod_2))
        #display(ReportText('Dzięki temu możemy korzystać z przygotowanych wcześniej klas i modułów'))
        display(ReportText('exemplary entry - fill as needed'))        

class PythonBasicsScheduleComponent(ReportComponent):

    #title="Klasa LibrariesImportComponent i jej zastosowanie"
    title="Schedule of course on Python Basics"

    def append_elements(self):
        classname = self.reported_object['classname']
        class_module = self.reported_object['module']
        class_field = self.reported_object['field']
        target = self.reported_object['target']
        #display(ReportText('Przed pracą w jupiterze należy zaimportować biblioteki'))
        display(ReportText('Before working in jupiter, import libraries'))
        display(ObjectCode(kod_2))
        #display(ReportText('Dzięki temu możemy korzystać z przygotowanych wcześniej klas i modułów'))
        display(ReportText('This allows us to use previously prepared classes and modules'))
        
        #display(ReportText('Przed pracą w jupiterze należy zaimportować biblioteki'))
        display(Markdown('''The course consists of following classes:

    -'zajecia_1': 'Wprowadzenie do przedmiotu, Przedstawienie guide z obsługi cocalca i zakładania jupytera, Syntax, typy zmiennych, przydzielenie tematów do prezentacji i rozpisanie issues - IntroToCocalc. (Tematy do prezentacji: Cocalc, Jupyter Notebook, import bibliotek - na przykłądnie sympy, numpy)'
    
    -'zajecia_2': "Podstawy gołego Pythona - podstawowe operacje na zmiennych, funkcje, klasy"
    
    -'zajecia_3': 'Biblioteki typu numpy, sympy itd. (studenkie propozycje przykladow zastoseowani z np prostej symulacji)',
    
    -'zajecia_4': 'Raportowanie zaawansowane (guide z raportowania)',
    
    -'zajecia_5': 'Odesystemy (przykładowe rozwiazanie równania różniczkowego i np. trajektoria)',
    
    -'zajecia_6': 'Pandas + symulacja a obiekcie pandas',
    
    -'zajecia_7': 'Wprowadzenia do modułu DynPy - helpy, guidy, tworzenie nowych systemów itd. gdzie szukać jak używać żeby pokazać co można finalnie stworzyć',
    
    -'zajecia_8': 'Budowanie komponentów i tutaj już jakiś zrobić np do systemu zrobionego wcześniej' ??,
    
    -'zajecia_9': 'Full raport na systemie i komponentach stworzony od zera (tutaj może już z tych trzech mini projekt w postaci zastowania pythona) - ???',
    
    -'zajecia_10': 'PROJEKT',
    
    -'zajecia_11': 'PROJEKT',
    
    -'zajecia_12': 'PROJEKT',
    
    -'zajecia_13': 'PROJEKT',
    
    -'zajecia_14': 'PROJEKT',
    
    -'zajecia_15': 'PROJEKT',

'''))

        display(ReportText('exemplary entry - fill as needed'))        

        
class PythonBasicsGuidelinesComponent(ReportComponent):

    #title="Klasa LibrariesImportComponent i jej zastosowanie"
    title="Main requirenents for Python Basics course"
        
    def append_elements(self):
#         classname = self.reported_object['classname']
#         class_module = self.reported_object['module']
#         class_field = self.reported_object['field']
#         target = self.reported_object['target']
        #display(ReportText('Przed pracą w jupiterze należy zaimportować biblioteki'))
        display(ReportText('Before working in jupiter, import libraries'))
        display(ObjectCode(kod_2))
        #display(ReportText('Dzięki temu możemy korzystać z przygotowanych wcześniej klas i modułów'))
        display(ReportText('This allows us to use previously prepared classes and modules'))
        zajecia = {}
        tematy = {}


        zajecia_list = [
            'Wprowadzenie do przedmiotu, cocalc biblioteki importy (na przykłądnie sympy nu,py) Kernel its (guide z obsługi cocalca i zakładania jupytera), raportowanie, przydzielenie prezenracja i rozpisanie issues - IntroToCocalc.',
            'Podstawy gołego pythona(prosty guide/raport z np importów i np min max, pętli.',
            'Biblioteki typu numpy sympy itd.(studenkie propozycje przyllakdow zastoseowani z np prostej symulacji).',
            'Raportowanie zaawansowane (guide z raportowanie ???)',
            'Odesystemy(przykładowe rozwiazanie rra i np trajektoria)',
            'Systemy dynamicczne tabelki(przedstawienie wyników obliczen z 6 i 7 w formie tabeli)',
            'Symulacje(pełny raport z symulacji)',
            'Przedstawienie obsługi dynpow typu helpy guidy its, gdzie szukać jak używać żeby pokazać co można finalnie stworzyć',
            'Budowanie systemów i tutaj już może coś zbudować',
            'Budowanie komponentów i tutaj już jakiś zrobić np do systemu zrobionego wcześniej',
            'Full raport na systemie i komponentach stworzony od zera#(tutaj może już z tych trzech mini projekt w postaci zastowania pythona'

    ]

        tematy_list = [
            'Celem pracy jest przygotowanie środowiska języka programowania Python do implementacji modeli związanych z mechaniką analityczną oraz klasyczną, z naciskiem na klasy związane z metodami małego parametru(metody pertubacyjne).',
            'Celem pracy jest przygotowanie środowiska języka programowania Python do implementacji modeli związanych z mechaniką analityczna oraz klasyczną, z naciskiem na opis różnych technik symulacji.',
            'Celem pracy jest zamodelowanie systemu autonomicznej jazdy w środowisku Python (aktywnego tempomatu) oraz zastosowanie Python’a jako narzędzia programistycznego do obliczeń inżynierskich.',
            'The objective of this paper is to examine the parametric vibration of a gearbox with variable mesh stiffness using the multiple time scale method. The study employs two linear models: a two-degree-of-freedom model for numerical analyses and a simplified one-degree-of-freedom gear model for deriving the analytical solution.',
            'Celem pracy jest rozpoznanie możliwości zastosowania języka programowania Python do analizy i badania układów dynamicznych oraz jako narzędzia dla inżynierów z wykorzystaniem biblioteki DynPy.',
            'Celem pracy jest przygotowanie klas wspomagających raportowanie w środowisku języka programowania Python. Stanowią one uzupełnienie do narzędzia umożliwiającego interaktywne rozwiązywanie zagadnień z zakresu mechaniki teoretycznej, z naciskiem na umożliwienie docelowemu użytkownikowi na uzyskanie szczegółowych informacji o metodach analizy.',
            'Celem pracy jest użycie języka programowania Python celem zbadania i analizy dynamiki ogniwa li-ion zastosowanego w modelu pojazdu elektrycznego.',
            'The aim of the thesis is to create model of theoretical battery in Python programming language, as module to Dynpy library.',
            'The thesis aims to conduct a comprehensive comparative analysis of various Battery Management System (BMS) models to assess their performance metrics using simulation techniques in Python. The study will focus on evaluating critical factors such as energy efficiency, state-of-charge estimation, battery lifespan maximization, and overall system reliability. It will also investigate the adaptability of these models to different types of battery chemistries and configurations. The goal is to identify the most effective BMS model that optimizes battery usage and extends the operational life of battery-powered systems.',
            'Celem pracy jest przygotowanie środowiska języka programowania Python do wspomagania prac projektowych z projektu Podstaw Konstrukcji Maszyn',
            'The aim of the paper is investigation of possiblities of utilisation of Python programming language in complex engineering calculations with an outcome in form of academic specialized text',
            'The  goal  of  the  paper  is  to  analyze  the  parametricvibration  of  a  gearbox  with  variable  mesh  stiffness  using  the  multiple  time  scale  method.  It  wasassumed  that  two  linear  models  will  be  used  for  the  study:  a  model  with  two  degrees  of  freedom(numerical  analyses)  and  a  reduced  gear  model  with  one  degree  of  freedom  (to  determine  the analytical solution).', 
            'Celem pracy jest użycie języka programowania Python do modelowania i analizy dynamiki pojazdów z uzgleniem wytrzymałości zmęczeniowej, zjawiska rezonansu oraz bilansu energetycznego'
    ]

        display(ReportText('''
        Python w zastosowaniach inżynierskich i naukowych

        - Przedmiot został stworzony na potrzeby nowej specjalności „Zaawansowane metody projektowania i rozwoju produktu w inżynierii mechanicznej” na studiach II stopnia (magisterskich). Odpowiada potrzebie opracowywania i zastosowania w praktyce innowacyjnych metod kształcenia opartych na pracy zespołowej i rozwiązywaniu problemów projektowych.

        - Idą było stworzenie i uruchamianie modułowych zajęć dydaktycznych o charakterze interdyscyplinarnym, które będą koncentrowały się na wykonywaniu zadań związanych z modelowaniem i rozwiązywaniem kompleksowych problemów inżynierskich oraz naukowych przy użyciu języka programowania Python.

        - Moduł zajęć dydaktyczny tworzą dwa powiązane przedmioty:
        1. Python w zastosowaniach inżynierskich i naukowych.
        2. Zaawansowane metody komputerowego modelowania maszyn i pojazdów.

        - Kompetencje zdobyte na przedmiocie Python w zastosowaniach inżynierskich i naukowych stanowią podstawę dla bardziej zaawansowanych zastosowań języka programowania Python, które realizowane są na przedmiocie Zaawansowane metody komputerowego modelowania maszyn i pojazdów.

        - Podczas zajęć studenci zapoznają się z metodykami programowania w paradygmacie obiektowym z wykorzystaniem mechanizmów języka Python. Opanowują składnię tego języka programowania w kontekście bogatego ekosystemu bibliotek i technologii, które przyczyniają się do jego dużej i wciąż rosnącej popularności. Jednocześnie zdobywają wiedzę z zakresu matematyki obliczeniowej: algorytmy, obliczenia symboliczne i numeryczne, przetwarzanie oraz analiza i wizualizacja danych.

        - Zadania będą realizowane zespołowo. Taka forma sprzyja rozwijaniu umiejętność pracy grupowej, tworzy przestrzeń do „burzy mózgów” i pozwala na efektywniejsze zdobywanie umiejętności projektowania i programowania obiektowego w kontekście rozważanego zagadnienia. Ułatwi to tworzenie adekwatnej architektury aplikacji z uwagi na zakres wymagań.

        - Tworząc przyjazną przestrzeń własnej pracy zespołowej, studenci nabędą umiejętności, które są przydatne podczas studiów, pracy naukowej, a przede wszystkim na współczesnym rynku pracy. Praktykują stosowanie technik i technologii programistycznych przedstawionych na zajęciach wykonując zadania związane z modelowaniem i rozwiązywaniem kompleksowych problemów.
        '''))

        display(ReportText('''
        Adaptacyjna forma zajęć:
        Jest to pierwsza edycja ww. zajęć, które będą poprowadzone na nowej specjalności studiów II st. magisterskich.
        - Wcześniej przeprowadziliśmy kilka względnie krótkich kursów „wdrożeniowych” dla studentów SiMR skupionych wokół grupy naukowo-dydaktycznej Effective Python, którą wspólnie założyliśmy i prowadzimy od 21.02.2022.
        - Tak więc, sami jesteśmy ciekawi jak to ostatecznie wyjdzie. Drodzy Studenci, wspólnie z nami będziecie kreować obraz zajęć w ramach tych dwóch przedmiotów modułowych. W dużej mierze to od Was, waszych chęci i zaangażowania będzie zależało to ile i jak uda nam się zrobić.
        - W ogólnym ujęciu zajęcia podzielone są na 2 fazy:
        I.	Szkolenie z podstaw programowania obiektowego z użyciem języka Python
        \(9 tygodni\)
        II.	Projekt (zespołowy) – praktyczne rozwiązanie zadanego/wybranego problemu
        \(6 tygodni\)
        '''))

        for i in range(11):
            zajecia[str(i + 1)] = zajecia_list[i]
            tematy[str(i + 1)] = tematy_list[i]

        df = pd.DataFrame(data={'Tematyka zajęć': zajecia, 'Guide': tematy})
        df1 = LatexDataFrame(df)


        display(df1.reported())
        display(Markdown('''
        FAZA I
        - Poszczególne zagadnienia przedmiotowe realizowane według harmonogramu zajęć.
        - Każde zajęcia podsumowuje raport konkretnego zastosowania Pythona.
        - Studenci prowadzą spotkania zajęciowe.
        Spokojnie, bez ciśnień, bez stresu. Chodzi o samokształcenie i swobodną formę zajęć, w ramach których na forum grupy każdy student przedstawia swoim kolegom/koleżankom wybrane zagadnienie podstawowe z harmonogramu, tzn. prezentuje i omawia dany guide, wyjaśnia prosto, jak korzystać. Tak jak to bywa w pracy na spotkaniach grupy projektowej czy programistycznej (sprinty Scrum).
        - Fazę I kończy spotkanie grupy, mające na celu podsumowanie dotychczasowych działań, wytyczenie kierunku dalszych prac, tzn. wydanie zadań do wykonania.
        '''))

        display(ReportText('''A przyświeca temu idea, którą określają trzy znane cytaty:

        “If you can't explain it simply, you don't understand it well enough”.
        (Jeśli nie potrafisz czegoś wyjaśnić prosto, nie rozumiesz tego wystarczająco dobrze.)
        Albert Einstein

        “What we don't understand we can make mean anything”.
        (To, czego nie rozumiemy, może mieć dowolne znaczenie.)
        Chuck Palahniuk
        A chodzi o to, żeby to miało właściwe znaczenie, żeby nie przeinaczasz sensu. ;-)

        “Knowing is not enough; we must apply. Willing is not enough; we must do”.
        (Wiedza nie wystarczy, musimy ją zastosować. Chęć nie wystarczy, musimy działać.
        Johann Wolfgang von Goethe'''))

        display(Markdown('''
        FAZA II
        - Planowanie i śledzenie postępu prac: zadania (issues) wyznaczone na GitHub. To jest hostingowy serwis internetowy przeznaczony do projektów programistycznych wspieranych przez system kontroli wersji Git.
        - Spotkania cotygodniowe poświęcone na weryfikację progresu prac (sprinty Scrum).
        - Wykonanie szablonowych raportów z zadań, według wytycznych.
        - Fazę II kończy spotkanie grupy, poświęcone podsumowaniu efektów prac: rozliczenie raportu i wykonanych zadań (issues).
        '''))
        
sample_code=('''
from dynpy.utilities.creators import ReserachIssueProjectCreator
ReserachIssueProjectCreator('Your issue name')
''')
class SampleComponent(ReportComponent):


    title="ReserachIssueProjectCreator class and its use"
        
    #def append_elements(self):
        #display(ReportText('Your text here'))
        #display(ObjectCode(sample_code))
    def append_elements(self):
        from .....utilities.creators import MeetingIssueCreator
        display(ReportText('This class is used to create meeting issues through API. The class consists of the following code: \\newline'))
        system = self.reported_object
        display(ObjectCode(sample_code))
        display(ReportText(' \\newline To learn more about this class use help. \\newline'))



from sympy import *
import sympy as sym
import itertools as itools
from pylatex import (Document, Package, Command
                     #Section, Subsection, Subsubsection, Itemize,  HorizontalSpace, Description, Marker
                    )
from pylatex.base_classes.containers import Container
#from pylatex.section import Paragraph, Chapter
from pylatex.utils import (#italic, 
                           NoEscape)

from ..adaptable import *
from ..report import *
from ..report import display as ip_display


def display(obj):

    #ip_display(IPMarkdown('=='*25 +'\n \n \n ==== if u see it let me (@bogumilchilinski) know' ))
    return ip_display(obj)


class MultivariableTaylorSeries(Expr):
    
    def __new__(cls,expr, variables,*args, n=2, x0=None):
        
        obj=super().__new__(cls,expr,variables,*args)
        obj._vars = variables
        obj._order = n
        obj._op_point = x0
        
        obj._expr_symbol = None
        
        return obj

    def _set_default_op_point(self):
        '''
        It sets 0 as op_point if x0 (look __new__) is not provided. For lack of op_point, the result is the same as MacLaurina series 
        '''
        
        if self._op_point is None:
            self._op_point = {coord:0 for coord in self._vars}
        
        return self._op_point
    
    def _args_shifted(self,*args):
        
        self._set_default_op_point()
        
        args_shifted = {
            arg: arg - arg_shift
            for arg, arg_shift in self._op_point.items()
        }
        
        return args_shifted

    def _diff_orders_dict(self):
            
        order_max = self._order
        args = self._vars
        args_shifted = self._args_shifted()
        
            
        diff_orders_list = sum([
            list(itools.combinations_with_replacement(args, order))
            for order in range(1, order_max + 1, 1)
        ], [])
        
        
        diff_orders_dict = {
            comp: (sym.Mul(*comp).subs(args_shifted) / sym.Mul(*[
                sym.factorial(elem) for elem in sym.Poly(sym.Mul(*comp), *args).terms()[0][0]
            ])).doit()
            for comp in diff_orders_list
        }

        return diff_orders_dict

    
    def _diff_symbols_dict(self):
        
        diff_orders_dict=self._diff_orders_dict()
        expr=self.args[0]
        op_point=self._op_point
        
        
        
        return {S.Zero:Subs(expr,list(op_point.keys()),list(op_point.values())),**{args_tmp:Subs(Derivative(expr,*args_tmp,evaluate=False),list(op_point.keys()),list(op_point.values())) 
            for args_tmp, poly in diff_orders_dict.items()}}

    def _diff_expr_dict(self):
        
        diff_orders_dict=self._diff_orders_dict()
        expr=self.args[0]
        op_point=self._op_point
        
        return {S.Zero:expr.subs(op_point),**{args_tmp:expr.diff(*args_tmp).subs(op_point)
            for args_tmp, poly in diff_orders_dict.items()}}
    
    def _components_dict(self):
        
        diff_orders_dict=self._diff_orders_dict()
        derivatives_dict=self._diff_symbols_dict()
        
        expr=self.args[0]
        op_point=self._op_point
        
        return {
                Subs(expr,list(op_point.keys()),list(op_point.values())):expr.subs(op_point),
                **{derivatives_dict[args_tmp] : expr.diff(*args_tmp).subs(op_point) for args_tmp, poly in diff_orders_dict.items()}
        }
        
    def _series(self):
        diff_orders_dict=self._diff_orders_dict()
        diff_dict=self._diff_symbols_dict()
        
        expr=self.args[0]
        op_point=self._op_point
        
        return expr.subs(op_point).doit()+Add(*[expr.diff(*args_tmp).subs(op_point).doit() * poly for args_tmp, poly in diff_orders_dict.items()],evaluate=False)
    
    def _symbolic_sum(self):
        diff_orders_dict=self._diff_orders_dict()
        diff_dict=self._diff_symbols_dict()
        
        expr=self.args[0]
        op_point=self._op_point
        
        return Subs(expr,list(op_point.keys()),list(op_point.values()))+Add(*[Mul(diff_dict[args_tmp]  ,poly,evaluate=True) for args_tmp, poly in diff_orders_dict.items()],evaluate=False)
    
    def _latex(self,*args):
        
        diff_orders_dict=self._diff_orders_dict()
        diff_dict=self._diff_symbols_dict()
        
        expr=self.args[0]
        op_point=self._op_point
        
        return '+'.join([latex(Mul(diff_dict[args_tmp]  ,poly,evaluate=True)) for args_tmp, poly in diff_orders_dict.items()])

    
    def calculation_steps(self,expr_symbol=None,form=None):

        obj = self
        
        if expr_symbol is None:
            obj._expr_symbol = self.args[0]
            
        obj_sym = self.__class__(expr_symbol,self._vars, n=self._order, x0=self._op_point)
        
        expr_dict=(self._diff_expr_dict())
        diffs_dict=(obj_sym._diff_symbols_dict())
        

        
        
        return [Eq(diffs_dict[key],expr_dict[key].doit())   for  key  in diffs_dict.keys()]
    
    def __str__(self,*args):
        return (self.args[0]).__str__()
    
    def __repr__(self,*args):
        return (self.args[0]).__repr__()

    
    
    
    def doit(self,*args):
        return self._series()


class ReportComponent(Subsection):

    latex_name = 'subsection'
    packages=[
              Package('standalone'),
              Package('siunitx')
             ]
    
    title='Report generic component'

    def __init__(self, reported_object, title=None, numbering=False, *, label=True, **kwargs):
        """
        Args
        ----
        title: str
            The section title.
        numbering: bool
            Add a number before the section title.
        label: Label or bool or str
            Can set a label manually or use a boolean to set
            preference between automatic or no label
        """



        self.reported_object = reported_object #it's forced by pylatex library
        
        
        if title is None:
            title = self.dynamic_title()
        
        
        
        super().__init__(title=title, numbering=numbering, label=label, **kwargs)
        CurrentContainer(self)
        
        if self.title is not None:
            ip_display(IPMarkdown(f'## {self.title}'))
        
        self.append_elements()
        
        
    def dynamic_title(self):
        return self.title
        
    def append_elements(self):
        pass

    def as_frame(self):
        frame=Frame(title=self.title,options=['allowframebreaks'])
        #frame.packages +(self.packages)
        frame+=(list(self))
        return frame

    @property
    def reported_object(self):

        from ...solvers.linear import ODESystem
        return self._reported_object       
#         if isinstance(self._reported_object, ODESystem):
#             return self._reported_object
#         else:
#             return self._reported_object._ode_system
            
    @reported_object.setter
    def reported_object(self, obj):
        self._reported_object=obj

    @property
    def _system(self):
        print('Kod do poprawienia #################### bo stosujesz starą zmienną self._system')
        print('zamień se self._system na self.reported_object')
        
        return self.reported_object
    
class ODESystemOperations(ReportComponent):
    
    def __init__(self, reported_object):
        from ...solvers.linear import ODESystem
        
        if isinstance(reported_object, ODESystem):
            ode_sys = reported_object
            ode_dummy = self.get_dummy()
            reported_object = (ode_sys, ode_dummy)

        super().__init__(reported_object)


    def get_dummy(self):
        from ...solvers.linear import ODESystem
        
        t = Symbol('t')
        m = Symbol('m',postive=True)
        g = Symbol('g',postive=True)
        c = Symbol('c',postive=True)
        b = Symbol('b',postive=True)
        x = Function('x')(t)
        
        return  ODESystem(odes=Matrix([m * x.diff(t,t)]), dvars=Matrix([x]), odes_rhs=Matrix([- c * x.diff(t) - b * x.diff(t,t) ], ode_order=2))
    
    def get_zero_ode_sys(self):
        from ...solvers.linear import ODESystem
        
        t = Symbol('t')
        x = Function('x')(t)
        return ODESystem(odes=0, dvars=x)

class TitlePageComponent(Environment):
    
    latex_name='titlepage'
    
    def __init__(self, system=None, options=None, arguments=None, start_arguments=None,
                 **kwargs):
        r"""
        Args
        ----
        options: str or list or  `~.Options`
            Options to be added to the ``\begin`` command
        arguments: str or list or `~.Arguments`
            Arguments to be added to the ``\begin`` command
        start_arguments: str or list or `~.Arguments`
            Arguments to be added before the options
        """

        self.system = system
        self.options = options
        self.arguments = arguments
        self.start_arguments = start_arguments

        
        
        super().__init__(options=options, arguments=arguments, start_arguments=start_arguments,**kwargs)
        
        if self.system is not None:

        
            system = self.system


            
            self.append(NoEscape('\centering'))

            self.append(NoEscape('\\Huge DRGANIA MECHANICZNE \n \n'))
            
            self.append(Command('vspace',arguments='1cm'))

            
            
            if len(system.q)==1:
                dof_str = 'JEDNYM STOPNIU SWOBODY'
            else:
                dof_str = 'WIELU STOPNIACH SWOBODY'
                
            if system._dissipative_potential==0 or system._dissipative_potential is None:
                damping_str = 'NIETŁUMIONE'
            else:
                damping_str = 'TŁUMIONE'

           
            self.append(NoEscape(f'\\Large {damping_str} UKŁADY O {dof_str} \n \n'))    
            
            self.append(Command('vspace',arguments='1cm'))
            
            self.append(NoEscape(f'{system._label} \n \n'))
            
            self.append(Command('vspace',arguments='1cm'))
            
            self.append(Command('MyAuthor'))
            self.append(NoEscape(f'\\par'))
            self.append(Command('vspace',arguments='1cm'))
            
            #self.append(Command('vspace',arguments='1cm'))
            #self.append(NoEscape(f'\\protect\\par'))
            #self.append(NewLine())
            self.append(Command('MyDate'))

 


    
class ExemplaryPictureComponent(ReportComponent):
    
    title="Przykład rzeczywistego obiektu"
    packages=[Package('float')] 
                                
    def append_elements(self):
        
        system = self._system

        display(ReportText(f'''Ilustracja przedstawia rzeczywisty obiekt mechaniczny, będący przedmiotem modelowania i analizy dynamicznej.
                            '''))

        with self.create(Figure(position='H')) as fig:
            fig.add_image(system._real_example(),width='8cm')

        display(ReportText(f'''Model dynamiczny układu określa się na podstawie analizy rozważanego przypadku. 
        Należy pamiętać, że stopień odwzorowania (poziom abstrakcji) modelu zależy od tego do czego planuje się go używać.
                            '''))            

        
class SchemeComponent(ExemplaryPictureComponent):
    title="Schemat układu"


    def append_elements(self):
        
        system = self._system

        display(ReportText(f'''Ilustracja przedstawia schemat rzeczywistego obiektu mechanicznego, wyznaczony na podstawie uprzedniej analizy rzeczywistego obiektu.
                            '''))
        
        with self.create(Figure(position='H')) as fig:
            fig.add_image(system._scheme(),width='8cm')

        display(ReportText(f'''Analizując przedstawiony układ można stwierdzić, że jego liczba stopni swobody to {len(system.q)}.
                            '''))



class NumericalAnalysisComponent(ExemplaryPictureComponent):
    title="Symulacja numeryczna"

    _default_ics = None
    _default_parameters = {}
    _default_tspan = np.linspace(0,1,1000)
    
    @classmethod
    def set_default_ics(cls,default_ics=None):
        
        if default_ics is None:
            cls._default_ics = default_ics
        
        return cls

    @classmethod
    def set_default_parameters(cls,default_parameters=None):
        
        if default_parameters is None:
            cls._default_parameters = default_parameters
        
        return cls
    
    @classmethod
    def set_default_tspan(cls,default_tspan=None):
        
        if default_tspan is None:
            cls._default_tspan = default_tspan
        
        return cls
    

    def append_elements(self):
        
        system = self._system

        display(ReportText(f'''Dla Damiana :P
                            '''))
        if self._default_ics is None:
            ics = list(system.Y*0.0)
        else:
            ics = self._default_ics
            
        sym_res = system.subs(self._default_parameters).numerized().compute_solution(self._default_tspan,ics)

        LatexDataFrame.formatted(sym_res).plotted(preview=True)
        

        display(ReportText(f'''Dla Damiana :P
                            '''))

class KineticEnergyComponent(ReportComponent):
    
    title="Energia kinetyczna"
    
    def append_elements(self):
        
        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys

        


        display(ReportText(f'''
                           Energia kinetyczna układu wyrażona jest wzorem:
                           
                           '''))
        

        display(SympyFormula( Eq(Symbol('T'),
                     dyn_sys_lin._kinetic_energy) , marker=None))
               
        display(ReportText(f'''
                           Wyznaczona wielkość określa energię układu wynikającą z jego własności inercyjnych (energię zmagazynowaną w elementach bezwładnych).
                           
                           '''))
    
    
    
class PotentialEnergyComponent(ReportComponent):
    
    title="Energia potencjalna"

    def append_elements(self):

        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys


        display(ReportText(f'''
                           Energia potencjalna układu wyrażona jest wzorem:
                           '''))

        display(SympyFormula( Eq(Symbol('V'),
                     dyn_sys_lin._potential_energy ), marker=None))

        display(ReportText(f'''
                           Zaprezentowana zależność opisuje oddziaływanie potencjalnych pól sił w których znajduje się obiekt.
                           '''))        
        
class DissipationComponent(ReportComponent):  # Marcel
    
    title="Dyssypacyjna funkcja Rayleigh'a"
    
    def append_elements(self):
        
        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys

        

        display(ReportText(f'''
                           Energia rozpraszana tłumieniem wyrażona jest wzorem:
                           
                           '''))

        display(SympyFormula( Eq(Symbol('D'),
             dyn_sys_lin._dissipative_potential) , marker=None ))

        display(ReportText(f'''
                           Podana zależność stanowi potencjał dysynpacyjny Rayleigh'a, 
                           który poddany różniczkowaniu względem wektora prędkości uogólnionych pozwala na określenie sił wiskotycznego tłumienia.
                           '''))
        
class SpringForce(ReportComponent):
    
    title="Siła od sprężyny"

    def append_elements(self):

        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys


        display(ReportText(f'''
                           Siła od sprężyny wyrażona jest wzorem:
                           '''))
        display(SympyFormula( Eq(Symbol('F'),-1*system._left_mount.stiffness*system.z,evaluate=False)))
        
        display(ReportText(f'''
                           zamiast x używam steady solution
                           '''))

        display(SympyFormula( Eq(Symbol('F'),-1*system._left_mount.stiffness*system.steady_solution(),evaluate=False)))


        display(ReportText(f'''
                           Siła od sprężyny, zwana również siłą naciągu, pojawia sie przy ściskaniu lub rozciaganiu. Siła, która działa jest przeciwnie skierowana do ruch i chce przywrócić do pierwotnego jej położenia. Zależy od sztywności sprężyny k oraz od tego o ile została rozciagnieta bądź skrócona x.
                           '''))
        
class DamperForce(ReportComponent):
    
    title="Siła tłumienia"

    def append_elements(self):

        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys


        display(ReportText(f'''
                           Siła tłumienia wyrażona jest wzorem:
                           '''))

        display(SympyFormula( Eq(Symbol('F'),-1*system.c*system.v,evaluate=False)))

        display(ReportText(f'''
                           zastępuje prędkość jako pochodna steady stolution po czasie:
                           '''))

        display(SympyFormula( Eq(Symbol('F'),-1*system.c*system.steady_solution().diff(system.ivar),evaluate=False)))

        display(ReportText(f'''
                           Siła tłumienia zmniejsza amplitude drgań, ma zwrot przeciwny do prędkości. Zależy od współczynnika tłumienia b oraz od prędkości v.
                           '''))                
class LogarithmicDecrement(ReportComponent):
    
    title="logarytmiczny dekrement tłumienia liczony z amplitud"

    def append_elements(self):

        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys


        display(ReportText(f'''
                           Logarytmiczny dekrement tłumienia wyrażona jest wzorem przy pomocy amplitud:
                           '''))

        display(SympyFormula( Eq(Symbol('delta'),log(system.A_n1/system.A_n2))))


        display(ReportText(f'''
                           Zaprezentowana zależność opisuje bezwymaiarową wielkość charakteryzującą intensywność tłumienia drgań swobodnych w przypadku podkrytycznym. Jest to wielkość stała dla rozpaywanego                                  układu drgającego i nie zależy od warrunków początkowych. Zależy natomiast od maksymalnych wychyleń w chwilach różniących się o okres drań tłumionych.
                           '''))
        display(ReportText(f'''
                           Logarytmiczny dekrement tłumienia u wyrażona jest wzorem przy pomocy okresu drgań tłumionych:
                           '''))
        display(SympyFormula( Eq(Symbol('delta'),2*pi*system.damping_coefficient()*(system.natural_frequencies())**-1,evaluate=False)))
        
        display(ReportText(f'''
                           Zaprezentowana zależność opisuje bezwymaiarową wielkość charakteryzującą intensywność tłumienia drgań swobodnych w przypadku podkrytycznym. Jest to wielkość stała dla rozpaywanego układu drgającego i nie zależy od warrunków początkowych.Zależy natomiast od okresu drań tłumionych i współczynnika h.
                           '''))


class LagrangianComponent(ReportComponent):
    
    title="Lagrangian (funkcja Lagrange'a)  układu"

    def append_elements(self):

        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys.linearized()

        
        print(system._scheme())
        
        mrk_lagrangian_nonlin=Marker('lagrangianNL',prefix='eq')

        #display(ReportText(f'''The following model is considered. The system's Lagrangian is described by the formula ({Ref(mrk_lagrangian_nonlin).dumps()}):
        #                    '''))
        display(ReportText(f'''Lagrangian systemu wyrażony jest wzorem ({AutoMarker(Eq(Symbol('L'),dyn_sys.L.expand()[0]))}):
                            '''))

        display((SympyFormula(  Eq(Symbol('L'),dyn_sys.L.expand()[0])  , marker=mrk_lagrangian_nonlin )  ))
        
        q_sym =[ Symbol(f'{coord}'[0:-3]) for coord in dyn_sys.q]
        
        diffL_d=lambda coord: Symbol(f'\\frac{{ \\partial L}}{{  \\partial {vlatex(coord)}  }}')
        diffD_d=lambda coord: Symbol(f'\\frac{{ \\partial D}}{{  \\partial {vlatex(diff(coord))}  }}')
        d_dt_diffL_d=lambda coord: Symbol(f'\\frac{{ \\mathrm{{d}}  }}{{  \\mathrm{{d}} {vlatex(system.ivar)}  }} {vlatex(diffL_d(coord))} ')        

        display(ReportText(f'''Równania Eulera-Lagrange'a dla rozważanego przypadku są nastęujące: 
                            '''))
        
        for coord in dyn_sys.q:
            display((SympyFormula(  Eq(d_dt_diffL_d(coord.diff(system.ivar)) - diffL_d(coord) + diffD_d(coord),Symbol(f'Q_{{ {vlatex(coord)} }}^N'))  , marker=mrk_lagrangian_nonlin,backend=vlatex )  ))
        
        
        display(ReportText(f'''Kolejne pochodne wynikające z zastosowania równań Eulera-Lagrange'a są nastęujące: 
                            '''))
        
        for coord in dyn_sys.Y:
            display((SympyFormula(  Eq(diffL_d(coord),dyn_sys.L.expand()[0].diff(coord))  , marker=mrk_lagrangian_nonlin,backend=vlatex )  ))
            


        for coord in dyn_sys.q.diff(system.ivar):
            display((SympyFormula(  Eq(d_dt_diffL_d(coord),dyn_sys.L.expand()[0].diff(coord).diff(system.ivar))  , marker=mrk_lagrangian_nonlin,backend=vlatex )  ))
        for coord in dyn_sys.q:
            display((SympyFormula(  Eq(diffD_d(coord),(S.One/2 * diff(dyn_sys.q.transpose())* dyn_sys_lin.damping_matrix()* diff(dyn_sys.q))[0].diff(diff(coord)))  , marker=mrk_lagrangian_nonlin,backend=vlatex )  ))
            #display(Markdown(f'\\begin{equation}    \\end{equation}').reported())
        #with doc_model.create(DMath()) as eq:
        #    eq.append(NoEscape(latex(Derivative(Symbol('L'),q_sym[0],evaluate=False))))
        #    eq.append(NoEscape('='))
        #    eq.append(NoEscape(vlatex(dyn_sys.L.expand()[0].diff(dyn_sys.q[0]))))

        mrk_gov_eq_nonlin=Marker('gov_eq_nonlin_sys',prefix='eq')

        #display(ReportText(f'''The governing equations of the system have a following form ({Ref(mrk_gov_eq_nonlin).dumps()}):
        #                    '''))

        display(ReportText(f'''Wyniki przedstawionych operacji wykorzystuje się wyznaczenia równań ruchu układu.
                            '''))


        
    def as_frame(self):
        frame=Frame(title=self.title,options=['allowframebreaks'])
        #frame.packages +(self.packages)
        frame+=(list(self))
        return frame
    
    
class GoverningEquationComponent(ReportComponent):
    
    title="Równania ruchu"
    
    def append_elements(self):
        
        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys.linearized()

        

        
        if len(system.q)==1:
            markers_str=f"przedstawia zależność ({AutoMarker(Eq(dyn_sys._eoms[0].simplify().expand(),0))})"            
        else:
            markers_str=f"przedstawiają zależności ({AutoMarker(Eq(dyn_sys._eoms[0].simplify().expand(),0))})-({AutoMarker(Eq(dyn_sys._eoms[-1].simplify().expand(),0))})"

        display(ReportText(f'''
                           Wykorzystując obliczone pochodne, wyznacza się równania ruchu na podstawie odpowiedniego wzoru.
                           Równania ruchu układu (nieliniowe w ogólnym przypadku) {markers_str}:
                           '''))
        
        for eq in dyn_sys._eoms:
            display(SympyFormula( Eq(eq.simplify().expand(),0) , marker=None))

        display(ReportText(f'''
                           Wyznaczone równania stanowią matematyczny opis dynamiczny właściwości układu.
                           Dalsza analiza pozwala na skuteczną analizę działania modelowanego obiektu i określenie jego parametrów mechanicznych.
                           '''))


class LinearizationComponent(ReportComponent):
    
    title="Linearyzacja równań ruchu"
    
    def append_elements(self):
        
        system = self._system
        ReportText.set_directory('./SDAresults')

        latex_store=AutoBreak.latex_backend
        AutoBreak.latex_backend = latex_store
        
        t=system.ivar
        

        dyn_sys=system
        dyn_sys_lin=dyn_sys.linearized()

        coords=tuple(list(dyn_sys.Y) + list(dyn_sys.q.diff(t,t)))
        op_point = {coord: 0 for coord in coords}

        #display(self._op_points(hint=hint, subs=True))
        op_point.update(dyn_sys._op_points(subs=True)[0])


        mrk_lagrangian_nonlin = Marker('lagrangLin',prefix='eq')
        mrk_lagrangian_lin = Marker('lagrangLin',prefix='eq')

        display(ReportText(
                f'''Linearyzaja równań polega na znalezieniu ich rozwinięcia w szereg Taylora względem współrzędnych, prędkości i przyspieszeń uogólnionych w otoczeniu punktu równowagi.
                Celem uproszczenia wprowadzono następujące oznaczenia:'''))

        for coord in coords:
            display((SympyFormula(  Eq(Symbol(vlatex(coord)),Symbol(latex(coord))) , marker=mrk_lagrangian_lin  )  )) 



        display(ReportText(
                f'''Punkty równowagi rozważanego układu są następujące:
                            '''))          

        for eq_coord,val in op_point.items():
            display((SympyFormula(  Eq(eq_coord,val) , marker=mrk_lagrangian_lin  )  ))


        diffL_d=lambda coord: Symbol(latex(Derivative(Symbol('L'),Symbol(vlatex(coord))))  )



        for no,eom in enumerate(dyn_sys._eoms):




            eq_sym=Symbol(f'RR_{latex(dyn_sys.q[no])}')


            display(ReportText(f'''Równanie ruchu dla współrzędnej ${latex(dyn_sys.q[no])}$ można przestawić jako:
                                '''))

            display((SympyFormula(  Eq(eq_sym,eom,evaluate=False) , marker=mrk_lagrangian_lin,backend=latex  )  ))


            display(ReportText(
                f'''Formalnie należy obliczyć pochodne cząstkowe wielkości uogólnionych ze składników równań Lagrange'a:
                            '''))


            display((SympyFormula(  Eq(MultivariableTaylorSeries(eq_sym,coords,n=1,x0=op_point)._symbolic_sum(),0) , marker=None,backend=latex  )  ))

            diff_list=MultivariableTaylorSeries(eom,coords,n=1,x0=op_point).calculation_steps(expr_symbol=eq_sym)

            display(ReportText(
                f'''Poszczególne pochodne mają następującą postać:
                            '''))

            for diff_eq in diff_list:

                display((SympyFormula(  diff_eq , marker=mrk_lagrangian_lin,backend=latex  )  ))

            display(ReportText(f'''Po podstawieniu obliczonych pochodnych, otrzumuje się następujące zlinearyzowane równanie:
                                '''))
            display((SympyFormula(  Eq(MultivariableTaylorSeries(eom,coords,n=1,x0=op_point).doit().expand().simplify().expand(),0,evaluate=False) , marker=mrk_lagrangian_lin,backend=latex  )  ))



        AutoBreak.latex_backend = latex_store


    
class FundamentalMatrixComponent(ReportComponent):
    
    title="Wyznaczanie macierzy fundamentalnej"
    
    def append_elements(self):
        
        system = self._system
        ReportText.set_directory('./SDAresults')

        latex_store=AutoBreak.latex_backend
        AutoBreak.latex_backend = latex_store
        
        t=system.ivar
        

        dyn_sys=system
        dyn_sys_lin=dyn_sys.linearized()


        display(ReportText(f'''Z równań ruchu wyznaczono macierz mas i sztywności układu:
                                '''))
        display((SympyFormula(  Eq(Symbol('M'),dyn_sys_lin.inertia_matrix(),evaluate=False) , marker='a' )  ))

        display((SympyFormula(  Eq(Symbol('K'),dyn_sys_lin.stiffness_matrix(),evaluate=False) , marker='a')  ))

        Delta = Symbol('\Delta')

        display(ReportText(f'''Macierz fundamentalna, na podstawie której wyznaczono równanie charakterystyczne rozważanego układu ${latex(Delta)}$, przedstawiają się następująco:
                                '''))

        display((SympyFormula(  Eq(Symbol('A'),dyn_sys_lin.fundamental_matrix(),evaluate=False) , marker='a'  )  ))
        display((SympyFormula(  Eq(Delta,dyn_sys_lin.fundamental_matrix().det().expand().simplify().simplify().expand(),evaluate=False) , marker='a',backend=latex  )  ))

        display(ReportText(f'''Macierz fundamentalna pozwala określić rozwiązanie ustalone. Natomiast bazując na równaniu charakterystycznym określa się częstości własne układu.
                                '''))

        AutoBreak.latex_backend = latex_store

class GeneralSolutionComponent(ReportComponent):
    
    title="Rozwiązanie ogólne"
    
    def append_elements(self):

        from ...dynamics import LagrangesDynamicSystem, HarmonicOscillator
        
        system = self._system
        ReportText.set_directory('./SDAresults')

        latex_store=AutoBreak.latex_backend
        AutoBreak.latex_backend = latex_store
        
        t=system.ivar
        

        dyn_sys=system
        dyn_sys_lin=dyn_sys.linearized()


        display(ReportText(f'''Rozwiązanie ogólne przedstawia wyrażenie:
                                '''))
        display((SympyFormula(  Eq(Symbol('X'),HarmonicOscillator(dyn_sys_lin.linearized(
                                            )).general_solution().n(3),
                                            evaluate=False) , marker='a',backend=latex  )  ))

        display(ReportText(f'''Rozwiązanie ogólne opisuje ruch analizowanego układu (przedstawia przemieszczenie w funkcji czasu) i wynika z rozważań dotyczących drgań swobodnych układu.
                                '''))

        AutoBreak.latex_backend = latex_store

        
        
class FrequencyResponseFunctionComponent(ReportComponent):
    
    title="Charakterystyka Amplitudowo-Częstotliwościowa"

    def append_elements(self):

        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys


        display(ReportText(f'''
                           funkcja odpowiedzi częstotliwościowej:
                           '''))

        display(SympyFormula( Eq(Symbol('frf'),
                     dyn_sys.frequency_response_function() ), marker=None))

        display(ReportText(f'''
                           jest to suma kwadratów amplitud pod pierwiastkiem
                           '''))  
        
FRFComponent = FrequencyResponseFunctionComponent

class FrequencyResponseFunctionComponentToSecond(ReportComponent):
    
    title="sens mocy układu"

    def append_elements(self):

        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys


        display(ReportText(f'''
                           sens mocy układu:
                           '''))

        display(SympyFormula( Eq(Symbol('frf^2'),
                     dyn_sys.frequency_response_function().doit()**2 ), marker=None))

        display(ReportText(f'''
                           Charakterystyka Amplitudowo-Częstotliwościowa podniesiona do kwadratu
                           '''))  
        


        

class SteadySolutionComponent(ReportComponent):
    
    title="Rozwiązanie szczególne"
    _phi=False
    def append_elements(self,phi=_phi):

        from ...dynamics import LagrangesDynamicSystem, HarmonicOscillator
        
        system = self._system
        ReportText.set_directory('./SDAresults')

        latex_store=AutoBreak.latex_backend
        AutoBreak.latex_backend = latex_store
        
        t=system.ivar
        

        dyn_sys=system
        dyn_sys_lin=dyn_sys.linearized()


        display(ReportText(f'''Rozwiązanie szczególne przedstawia wyrażenie:
                                '''))

        display((SympyFormula(  Eq(Symbol('X_s'),
                            HarmonicOscillator(dyn_sys_lin.linearized(
                            )).steady_solution().n(3),
                            evaluate=False) , marker='b',backend=latex  )  ))

        AutoBreak.latex_backend = latex_store

        display(ReportText(f'''Rozwiązanie szczególne związane jest obecnością wielkości wymuszających ruch (drgania) analizowanego układu.
                                '''))
        
class MaxStaticForce(ReportComponent):
    
    title="Maksymalna siła statyczna"

    def append_elements(self):

        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys


        display(ReportText(f'''
                           Wartość maksymalna siły statycznej działającej na pojedynczy element mocujący:
                           '''))

        display(SympyFormula( Eq(Symbol('F_s'),
                     dyn_sys.max_static_force().doit() ), marker=None))

        display(ReportText(f'''
                           Wartość maksymalna siły statycznej działającej na pojedynczy element mocujący.
                           '''))  
        
class MaxDynamicForce(ReportComponent):
    
    title="Maksymalna siła statyczna"

    def append_elements(self):

        system = self._system
        dyn_sys=system
        dyn_sys_lin = dyn_sys


        display(ReportText(f'''
                           Wartość maksymalna siły dynamicznej działającej na pojedynczy element mocujący:
                           '''))

        display(SympyFormula( Eq(Symbol('F_d'),
                     dyn_sys.max_dynamic_force().doit() ), marker=None))

        display(ReportText(f'''
                           Wartość maksymalna siły dynamicznej działającej na pojedynczy element mocujący:
                           ''')) 


from sympy import (Symbol, symbols, Matrix, sin, cos, diff, sqrt, S, diag, Eq,
                   Function, lambdify, factorial, solve, Dict, Number, N, Add,
                   Mul, expand,zoo,exp,Dummy, det)

from sympy.physics.mechanics import dynamicsymbols
from sympy.physics.vector.printing import vpprint, vlatex
import sympy as sym
from sympy.utilities.autowrap import autowrap, ufuncify
import numpy as np
import itertools as itools
import copy
import scipy.integrate as solver
from ..utilities.timeseries import DataMethods, SpectralMethods, TimeDomainMethods, SpectrumSeries, SpectrumFrame, TimeSeries, TimeDataFrame

from collections import ChainMap
from IPython.display import display
import sympy.physics.mechanics as me

from sympy.simplify.fu import TR8, TR10, TR7, TR3, TR0

from .linear import LinearODESolution, FirstOrderODE, AnalyticalSolution, FirstOrderLinearODESystem, FirstOrderODESystem, ODESystem, ODESolution,FirstOrderLinearODESystemWithHarmonics
from ..utilities.components.ode import en as ode_comp

#from timer import timer
from time import time
from collections.abc import Iterable
from functools import cached_property,lru_cache

from .tools import CommonFactorDetector, ODE_COMPONENTS_LIST, CodeFlowLogger

def const_no_gen():
    num = 0
    while True:
        yield num
        num += 1

class SimplifiedExpr:
    
    """

    This class provides methods which allow to obtain a differential equation in simplified form, or extract desired simplification parameters. 

    """

    _const_no=const_no_gen()
    
    _subs_container={}
    def __init__(self,expr,ivar,parameters = None ):

        self._expr = expr
        self._ivar = ivar
        self._parameters = parameters

    @property
    def _fun_list(self):
        
        """
        
        Method extracts functions from expression and returns them in a list.
        
        """
        
        funlist=[expr  for  expr in  (self._expr).atoms(sin,cos) if expr.has(self._ivar)]
        
        return funlist
    
    @property
    def simplified_expr(self):
        
        """
        
        This method returns a simplified expression, prepared in sum_expr method.
        
        """
        
        return self.sum_expr
    @property
    def sum_expr(self):
        
        """
        
        This method provides simplified expression that consist of main part of an equation and functions with additional coefficents.
        
        """

        data = self._expr_collector
        expr_dict={key:values[0]  for  key,values  in  data.items() }
        self.__class__._subs_container={**self.__class__._subs_container, **expr_dict}
        
        simplified_sum=sum([key*values[1]  for  key,values  in  data.items()  ],S.Zero)
        expanded_sum=sum([values[0]*values[1]  for  key,values  in  data.items()  ],S.Zero)
        
        expr_rest = (self._expr - expanded_sum).expand()
        # display(expr_rest)
        
        # print(len(self._expr.args))
        # print(len(simplified_sum.args))
        return simplified_sum + expr_rest

    
    @property
    def _expr_dict(self):
        
        """

        Method returns a dictionary that contains additional coefficients with their values.

        """
        
        expr_dict={key:values[0]  for  key,values  in  self._expr_collector.items() }
        return expr_dict
    @property
    def _expr_collector(self):
        
        """

        This method combines functions and coefficients and then returns a dictionary with segregated values.

        """
        
        parameters = self._parameters
        
        if parameters:
            aux_symbol = lambda no: Function(f'DummyFun_{next(self._const_no)}')(*parameters)
        else:
            aux_symbol = lambda no: Dummy(f'D_{no}')
        
        bucket = {aux_symbol(no):(self._expr.coeff(fun),fun) for no,fun in enumerate(self._fun_list)}
        return bucket
    
    @property
    def full_expr(self):
        
        """

        Main method that returns a full expression with functions and proper coefficients.

        """
        
        #display(self.__class__._subs_container)
    
        return self._expr.doit().expand().subs(self.__class__._subs_container)
    
    def __repr__(self,*args):
        return 'instance of SimplifiedExpr'

class PerturbationODESolution(ODESolution):
    _eps = Symbol('varepsilon')
    
    def set_small_parameter(self,eps,inplace = False):
        obj = copy.copy(self)
        obj._eps = eps
        
        return obj
    
    @property
    def small_parameter(self):
        return self._eps

    @small_parameter.setter
    def small_parameter(self,eps):
        
        self._eps = eps
        
        return self._eps
    
    @property
    def eps(self):
        return self.small_parameter
    
    def _calculate_constant(self,ics=None,*args,**kwargs):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """

        const_eqns=self._get_constant_eqns(ics).subs(self.eps,0).doit()
        const_list = self._spot_constant()
        
        #display('ivar',self.ivar)
        
        return solve(const_eqns,const_list)
    

class ODESystemApproximation(ODESystem):

    _simp_dict = {}
    _callback_dict = {}

    @property
    def _default_solver(self):
        
        return FirstOrderLinearODESystemWithHarmonics
        #return FirstOrderLinearODESystem
    
class NthOrderODEsApproximation(ODESystem):
    _ode_order=2

    _simp_dict = {}
    _callback_dict = {}
    
    # _const_list = []
    
    @property
    def _default_solver(self):
        
        return FirstOrderLinearODESystemWithHarmonics
        #return FirstOrderLinearODESystem
    
    @property
    def _secular_funcs(self):

        # weird error occurs, when .remove_secular_terms is called. if secular terms are missing, method returns None,
        # if statement should be considered
        
        #const_to_add=[eqn for eqn in self._const_list ]
        
        if self._parameters is None:
            params = []
        else:
            params = self._parameters
        
        secular_funcs = set() | (self.general_solution.rhs.atoms(Function) - set(
            self._const_list) - set(params) - {self.ivar})
        #display(secular_funcs)
        
        return {sec_fun for sec_fun in secular_funcs if sec_fun.has(self.ivar)}
    
    @property
    def _secular_conditions(self):

        sec_funcs = self._secular_funcs

        sec_conditions = Matrix(list(set(sum([list(self.odes.applyfunc(lambda entry: entry.coeff(func))) for func in sec_funcs],[]))-{0}))

        return sec_conditions

    @property
    def secular_terms(self):

        #print('secsec')
        sec_funcs = self._secular_funcs
        #self._const_list

        ##display('sec_funcs',sec_funcs)
        CodeFlowLogger(sec_funcs,'sec_funcs',self)
        
        sec_conditions = self._secular_conditions
        #display('sec_conditions',sec_conditions)
        ivar = self._parameters[0]
        
        #sec_conditions =FirstOrderODESystem.from_ode_system(ODESystem([sec_conditions[1][1],sec_conditions[3][1]],dvars = self._const_list ,ivar=ivar  )).linearized().solution
        
        # print('some check fof sec eq')
        

        CodeFlowLogger(sec_conditions,'sec_conditions',self)
        
        const_list = [fun_cons for fun_cons in list(sec_conditions.atoms(Function)) if  'C' in str(fun_cons) ]
        

        CodeFlowLogger(const_list,'const_list',self)
        
        sec_odes=ODESystem((sec_conditions),
                            dvars=Matrix(const_list),
                            ivar=ivar,  
                            ode_order=None)
        

        CodeFlowLogger(sec_odes.as_type(FirstOrderLinearODESystem).solution,'sec_odes.as_type(FirstOrderLinearODESystem).solution',self)
        #fode_harm_rhs = FirstOrderLinearODESystemWithHarmonics.from_ode_system(sec_odes)
        
        # print('transformed')
        # display(fode_harm_rhs)
        # display(type(fode_harm_rhs))
        
        return sec_odes

    def remove_secular_terms(self):


        obj = ODESystem.from_ode_system(self)

        # subs_dict = {comp: 0 for comp in self._secular_funcs}
        
        # return obj.subs(subs_dict).doit().subs(zoo,0).doit()
        
        def eqns_map(obj):

            oper_expr = lambda obj: (TR10(
                TR8(TR10(TR0(obj.expand())).expand())).expand().doit().expand())

            #oper_expr = lambda obj: TR8(obj).expand()

            #oper_expr = lambda obj: obj.doit().expand()

            if isinstance(obj, Add):
                elems = obj.args
                return sum(oper_expr(elem) for elem in elems)
            else:
                return oper_expr(obj)
            

        obj = obj.applyfunc(eqns_map)
        
        secular_expr_list = [obj.applyfunc(lambda row: row.coeff(comp)*comp).rhs  for comp in self._secular_funcs]
        secular_expr = sum(secular_expr_list,Matrix( [0]*len(obj) ))

        return ODESystem((obj.as_matrix() - secular_expr).expand().subs({comp:0 for comp in self._secular_funcs}).doit(),obj.dvars,ivar=obj.ivar)



        
#     def find_approximation(self):

#         zeroth_ord = self.odes.applyfunc(lambda ode: ode.general_solution.self.remove_secular_terms())

#         for order,approx in enumerate(self.odes):

#             approx._parameters = self.t_list[order+1:]

#             nth_ord = approx.applyfunc(lambda ode: ode.steady_solution.self.remove_secular_terms())

#         return {order:}




class MultiTimeScaleSolution(ODESystem):
    _ode_order=2
    _stored_solution = None
    _saved_solution = {}

    _saved_numerized = {}
    _order = 2
    _scales_no = None

    def __new__(cls,
                odes_system,
                dvars,
                ivar=None,
                ics=None,
                eps=Symbol('varepsilon'),
                omega=None,
                order=2,
                scales_no=None,
                t_span=[],
                params=[],
                params_values={},
                extra_params={},
                ic_point={},
                equation_type=None,
                label=None,
                evaluate=True,
                **options):

        if not isinstance(odes_system, Iterable):
            odes_system = Matrix([odes_system])
        if not isinstance(dvars, Iterable):
            dvars = Matrix([dvars])

        obj = super().__new__(cls,
                              odes_system,
                              dvars=dvars,
                              ivar=ivar,
                              evaluate=evaluate,
                              **options)

        obj._const_list = []

        if ivar is not None:
            obj._ivar = ivar

        if label == None:
            label = obj._label = obj.__class__.__name__ + ' with ' + str(
                len(obj.dvars)) + ' equations'
            
        obj._label = 123
                
        obj.eps = eps
        obj._order = order
        
        if scales_no is None:
            obj._scales_no = obj.order+1

        obj._stored_solution = None
        obj._saved_solution = {}
        obj.extra_params = extra_params

        obj.secular_eq = {}
        obj.int_const = set()

        obj.omega = 1
        if omega:
            obj.omega = omega
        
        obj.ics = ics

        obj._const_sol = {}

        
        obj._eoms = odes_system
        obj.odes_system = odes_system
        #obj._odes = odes_system

        return obj
    
    @property
    def eps(self):
        return self._eps
    
    @eps.setter
    def eps(self,eps):
        self._eps=eps
        
    @property
    def order(self):
        return self._order
    
    @property
    def scales_no(self):
        if self._scales_no is None:
            return self.order+1
        else:
            return self._scales_no
    
    @scales_no.setter
    def scales_no(self,no):
        
        self._scales_no = no
    
    @property
    def omega(self):
        return self._omega

    @omega.setter
    def omega(self, omega):
        self._omega = omega
        
    @property
    def ics(self):
        return self._ics

    @ics.setter
    def ics(self, ics):
        self._ics = ics
    @property
    def extra_params(self):
        return self._extra_params

    @ics.setter
    def extra_params(self, extra_params):
        self._ics = extra_params
        
    def __str__(self):
        return '123'

    def __repr__(self):
        return '123'

    def set_solution_order(self, order=1):

        self._order = order

    @property
    def _report_components(self):
        
        comp_list=[
        ode_comp.ODESystemComponent,
        ode_comp.VariablesComponent,

        ode_comp.ODESystemCodeComponent,
        ode_comp.MSMCalculationsOrderComponent,        
        ode_comp.PredictedSolutionComponent,
        ode_comp.DetailsOfPredictedSolutionComponent,

        ode_comp.GoverningEquationComponent,

        #ode_comp.ApproximatedNonlinearGoverningEquationComponent,
#         ode_comp.FundamentalMatrixComponent,
        ode_comp.ZerothOrderApproximatedEqComponent,
        ode_comp.FirstOrderApproximatedEqComponent,

        # ode_comp.GeneralSolutionComponent,
        ode_comp.ZerothOrderSolutionComponent,
        ode_comp.SecularTermsEquationsComponent,
            
        ]
        
        return comp_list


    
    @property
    def order(self):
        return self._order

    @order.setter
    def order(self,order):
        self._order = order

    def set_order(self,order):
        
        
        new_msm = MultiTimeScaleSolution(self.as_matrix(),self.vars,ivar=self.ivar,omega=S.One,order=order,scales_no=self.scales_no,eps=self.eps )
        new_msm._scales_no = self._scales_no
        
        return new_msm
    
    def _assign_properties(self,obj):
        
        obj._eps=self.eps     
        obj._vars = self._vars
        #obj._const_list = self._const_list
        
        return obj
    
    
    @property
    def ics_dvars(self):

        q = list(self.dvars)
        dq = list(Matrix(self.dvars).diff(self.ivar))

        return Matrix(q + dq)

#     @order.setter
#     def order(self, order):
#         self._order = order

    @cached_property
    def t_list(self):
        r"""
        Returns the list of time domanin slow varying function
        """

        self._t_list = [
            t_i(self.ivar)
            for t_i in symbols(f't_0:{self.scales_no}', cls=Function, real=True)
        ]

        return self._t_list
    
    @cached_property
    def _scales_formula(self):
        r"""
        Returns the list of time domanin slow varying function
        """

        t_list = self.t_list
        expr_list = [self.ivar*self.eps**no for no in range(len(t_list))]

        return AnalyticalSolution.from_vars_and_rhs(t_list, expr_list)
  
    def approx_function_symbol(self, order=0):
        dvars_no = len(self.dvars)

        if dvars_no == 1:
            func_str = str(self.vars[0]).replace(f"({self.ivar})", "")
        else:
            func_str = 'Y'

        return func_str + f"_{{{order}}}"

    def approximation_function(self, order, max_order=1, ivar_list=None):
        dvars_no = len(self.vars)

        if ivar_list is None:
            t_list = self.t_list
        else:
            t_list = ivar_list

        fun_str = self.approx_function_symbol(order)

        if dvars_no == 1:
            approx_fun = [
                sym.Function(fun_str)(*t_list)
                for dvar_no, dvar in enumerate(self.vars)
            ]
        else:
            approx_fun = [
                sym.Function(f"{fun_str[:-1]}{dvar_no}"+'}')(*t_list)
                for dvar_no, dvar in enumerate(self.vars)
            ]

        return Matrix(approx_fun)

    
#     def approx_function_symbol(self,order=0):

#         dvars_no = len(self.dvars)

#         if dvars_no==1:

#             func_str = '\\'+str(self.vars[0]).replace(f"({self.ivar})","")
#         else:
#             func_str = 'Y'
        
#         return func_str + f"_{{{str(order)}}}"

#     def approximation_function(self, order, max_order=1, ivar_list = None ):

#         dvars_no = len(self.vars)

#         if ivar_list is None:
#             t_list = self.t_list
#         else:
#             t_list = ivar_list
        
#         fun_str = self.approx_function_symbol(order)
        
#         if dvars_no==1:

#             aprrox_fun = [
#                 sym.Function(fun_str)(*t_list)
#                 for dvar_no, dvar in enumerate(self.vars)
#             ]
#         else:

#             aprrox_fun = [
#                 sym.Function(fun_str +  str(dvar_no) + '}')(*t_list)
#                 for dvar_no, dvar in enumerate(self.vars)
#             ]


# #             aprrox_fun = [
# #                 sym.Function(fun_str[0:-1] +  str(dvar_no) + '}')(*t_list)
# #                 for dvar_no, dvar in enumerate(self.vars)
# #             ]

#         return Matrix(aprrox_fun)

    def approximation_function_without_scales(self, order, max_order=1):

        return self.approximation_function(order=order,max_order=max_order,ivar_list=[self.ivar])
    
    def part_derivative(self, order=1):
        return Eq(self.approximation_function_without_scales(order).diff(self.ivar)[0],self.approximation_function(order).diff(self.ivar)[0])
    
    def sec_part_derivative(self, order=1):
        return Eq(self.approximation_function_without_scales(order).diff(self.ivar,2)[0],self.approximation_function(order).diff(self.ivar,2)[0])
    
    def part_derivative_subs(self, order=1):
        deriv_dict = {}

        for index, value in enumerate(self._scales_formula):
            scales_lhs = self._scales_formula.lhs[index]
            scales_rhs = self._scales_formula.rhs[index]

            scales_lhs_deriv = self.first_order_subs().lhs[index]
            scales_rhs_deriv = self.first_order_subs().rhs[index]

            scales_lhs_second_deriv = self.second_order_subs().lhs[index]
            scales_rhs_second_deriv = self.second_order_subs().lhs[index]

            deriv_dict[scales_lhs] = scales_rhs
            deriv_dict[scales_lhs_deriv] = scales_rhs_deriv
            deriv_dict[scales_lhs_second_deriv] = scales_rhs_second_deriv

        return Eq(self.approximation_function_without_scales(order).diff(self.ivar)[0],self.part_derivative(order).rhs.subs(deriv_dict).simplify().expand())
    
    def sec_part_derivative_subs(self, order=1):
        deriv_dict = {}

        for index, value in enumerate(self._scales_formula):
            scales_lhs = self._scales_formula.lhs[index]
            scales_rhs = self._scales_formula.rhs[index]

            scales_lhs_deriv = self.first_order_subs().lhs[index]
            scales_rhs_deriv = self.first_order_subs().rhs[index]

            scales_lhs_second_deriv = self.second_order_subs().lhs[index]
            scales_rhs_second_deriv = self.second_order_subs().lhs[index]

            deriv_dict[scales_lhs] = scales_rhs
            deriv_dict[scales_lhs_deriv] = scales_rhs_deriv
            deriv_dict[scales_lhs_second_deriv] = scales_rhs_second_deriv

        return Eq(self.approximation_function_without_scales(order).diff(self.ivar)[0],self.sec_part_derivative(order).rhs.subs(deriv_dict).simplify().expand())

    
    def predicted_solution(self, order=1, dict=False, equation=False):

        dvars_no = len(self.vars)
       # order = self.order

        solution = sum(
            (self.approximation_function(comp_ord, order) * self.eps**comp_ord
            for comp_ord in range(order + 1)), sym.zeros(dvars_no, 1))

        return ODESolution.from_vars_and_rhs(self.vars, solution)

    def predicted_solution_without_scales(self, order=1, dict=False, equation=False):

        dvars_no = len(self.dvars)

        solution = sum(
             (self.approximation_function_without_scales(comp_ord, order) * self.eps**comp_ord
             for comp_ord in range(order + 1)), sym.zeros(dvars_no, 1))

        return ODESolution.from_vars_and_rhs(self.dvars, solution)
    
    def derivative_without_scale(self, degree, order=2, dict=False, equation=False):
        
        derivative_rhs = self.predicted_solution_without_scales(self.order).as_eq_list()[0].n(3).rhs.diff(self.ivar,degree)
        derivative_lhs = self.dvars[0].diff(self.ivar,degree)
        
        return AnalyticalSolution(derivative_lhs, rhs = derivative_rhs, vars = self.dvars[0])

    def first_order_subs(self):

        first_ord_subs = {
            t_i.diff(self.ivar): self.eps**t_ord
            for t_ord, t_i in enumerate(self.t_list)
        }

        return AnalyticalSolution.from_dict(first_ord_subs)
    
    def second_order_subs(self):

        sec_ord_subs = {
            t_i.diff(self.ivar, 2): 0
            for t_ord, t_i in enumerate(self.t_list)
        }

        return AnalyticalSolution.from_dict(sec_ord_subs)
    
    #@lru_cache
    def eoms_approximation(self, order=None, odes_system=None):

        
        order = self.order
        
        if not odes_system:
            odes_system = self.as_matrix()

        first_ord_subs = self.first_order_subs().as_explicit_dict()
        sec_ord_subs = self.second_order_subs().as_explicit_dict()

        #display(self.predicted_solution(order).as_dict().subs(sec_ord_subs).doit().subs(first_ord_subs).doit())

        
        odes_rhs = self.as_matrix().subs(self.predicted_solution(
            order).as_explicit_dict()).subs(sec_ord_subs).doit().subs(first_ord_subs).doit()

        #display()
        #eoms_approximated = ODESystem(odes_system,dvars=self.dvars,parameters = self.t_list).subs(
        #    self.predicted_solution(order).as_dict())
        eoms_approximated = - odes_rhs

        eoms_approximated = eoms_approximated.subs(sec_ord_subs).doit().subs(
            first_ord_subs).doit()

        t_fun_subs_dict = ({
            expr_tmp: expr_tmp.subs(self.ivar, self.t_list[0])
            for expr_tmp in (
                eoms_approximated.atoms(Function) - {*self.t_list} - {
                    *sum([
                        list(self.approximation_function(ord_tmp, order))
                        for ord_tmp in range(order + 1)
                    ], [])
                })
        })

        CodeFlowLogger(eoms_approximated,'eoms for approximation before subs',self)
        CodeFlowLogger(t_fun_subs_dict,'dict with time subs',self)

        return eoms_approximated.subs(t_fun_subs_dict).doit()
    
    def eoms_approximation_as_eq_list(self):
        defined_eoms=self.eoms_approximation()
        approx_eoms=[Eq(-eap,0) for eap in defined_eoms]
        return approx_eoms

    #@lru_cache
    def eoms_approximation_list(self,
                                max_order=3,
                                min_order=0,
                                odes_system=None):

        eoms_approximated = self.eoms_approximation(
            order=max_order, odes_system=odes_system).expand()

        order_get = lambda obj, order: (obj.diff(self.eps, order) / factorial(
            order)).subs(self.eps, 0).doit()

        #NthOrderODEsApproximation
        
        
        approx_list=[
            NthOrderODEsApproximation(
                eoms_approximated.applyfunc(lambda obj: order_get(obj, order)),
                dvars=self.approximation_function(order),
                ivar=self.t_list[0],
                ode_order=2,
                parameters=self.t_list)
            for order in range(min_order, max_order + 1)
        ]
        
        
        return approx_list
        #return [NthOrderODEsApproximation.from_ode_system(approx_ode)   for approx_ode in approx_list]

    #@lru_cache
    def eoms_approx_with_const(self, order=1):

        self.secular_eq = {}

        approx_eoms_list = self.eoms_approximation_list(self.order)

        approx_eoms_list[0]._parameters = self._t_list[1:]
        
        # display(approx_eoms_list)
        # display(approx_eoms_list[0])
        
        sol = approx_eoms_list[0].solution
        # print('spot_const')
        # display(sol._spot_constant())
        
        #### !!!! 
        sol_list = [sol]
        
        
        sol_subs_list = [sol.applyfunc(lambda row: SimplifiedExpr(row,ivar=self._t_list[0],parameters=self._t_list[1:]).full_expr).doit() for sol in sol_list]
        sol_subs_dict  = {  dvar:eqn     for  sol   in sol_subs_list  for dvar, eqn in sol.as_explicit_dict().items()}

        
        # print('_gen_sol')
        # display(sol_list)
        # display(sol_subs_dict)
        
        # display(approx_eoms_list[0]._const_list)
        approx_with_const = [approx_eoms_list[0]]

        for order, approx in enumerate(approx_eoms_list[1:]):

            approx._parameters = self._t_list[1:]
            #approx = approx.as_first_ode_linear_system()

            eqns_map = lambda obj: (TR10(TR8(TR10(obj.expand()).expand())).
                                    expand().doit().expand())

            def eqns_map(obj):

                oper_expr = lambda obj: (TR10(
                    TR8(TR10(TR0(obj.expand())).expand())).expand().doit().expand())

                #oper_expr = lambda obj: TR8(obj).expand()

                #oper_expr = lambda obj: obj.doit().expand()

                if isinstance(obj, Add):
                    elems = obj.args
                    return sum(oper_expr(elem) for elem in elems)
                else:
                    return oper_expr(obj)
            
            # print('approx eqn')
            # display(approx)
            
            sol_subs_list = [sol.applyfunc(lambda row: SimplifiedExpr(row,ivar=self._t_list[0],parameters=self._t_list[1:]).full_expr).doit() for sol in sol_list]
            sol_subs_dict  = {  dvar:eqn     for  sol   in sol_subs_list  for dvar, eqn in sol.as_explicit_dict().items()}
            
            sol_subs_dict  = sol.as_explicit_dict()
        
            approx_subs = approx.applyfunc(eqns_map).subs(
                sol_subs_dict).applyfunc(eqns_map)
            
            # display(approx_subs)

            approx_subs = NthOrderODEsApproximation.from_ode_system(approx_subs)
            approx_subs._parameters = self.t_list[1:]
            
            approx_with_const += [approx_subs]
            
        return approx_with_const


    def _general_sol(self, order=1):

        self.secular_eq = {}

        
        approx_with_const = self.eoms_approx_with_const(order)

        sol = approx_with_const[0].solution
        sol_list = [sol]

        for order, approx in enumerate(approx_with_const[1:]):

            approx_subs = approx
            self.secular_eq[self.eps**(order +1)] = approx_subs.secular_terms#.as_type(FirstOrderLinearODESystem)
            
            #nonlin_ode.eoms_approx_with_const()[0].remove_secular_terms().steady_solution.rhs
            approx_subs = approx_subs#.steady_solution.rhs
            
            #aux_mat=approx_subs.jacobian(approx_subs.dvars)
            # if det(aux_mat)==0:

            #     SolverClass = FirstOrderLinearODESystem
            # else:
            #     SolverClass = FirstOrderLinearODESystemWithHarmonics

            # SolverClass = FirstOrderLinearODESystemWithHarmonics
            # #SolverClass = FirstOrderLinearODESystem
            
            if len(sol) > 0:
                ode_2_solve = approx_subs.subs(sol_list[-1].as_explicit_dict()).remove_secular_terms()
            else:
                ode_2_solve = approx_subs.remove_secular_terms()
            
            # ode_2_solve._ivar = self.t_list[0]
            
            # ode_2_solve = SolverClass.from_ode_system(ode_2_solve)

            

            sol = ode_2_solve.steady_solution.applyfunc(
                lambda obj: obj.expand())#.applyfunc(eqns_map)#.applyfunc(lambda row: SimplifiedExpr(row,ivar=self._t_list[0],parameters=self._t_list[1:]).sum_expr)


            sol_list += [sol.applyfunc(lambda row: SimplifiedExpr(row,ivar=self._t_list[0],parameters=self._t_list[1:]).full_expr).doit()]
            #sol_list += [sol]

        CodeFlowLogger(sol_list,'sol list',self)

        return sol_list


    @property
    def general_solution_with_consts(self):

        
        order = self.order
        
        #print('my order',order)
        sol_list = []
        gen_sol = self._general_sol(order)
        
        # print('gen_sol')
        # display(*gen_sol)
        
        if self.eps in self.secular_eq:

            ode2check = self.secular_eq[self.eps]#.as_first_ode_linear_system()

            
            CodeFlowLogger(ode2check,'sec_eq',self)
            
            # display(ode2check._as_fode())
            # display(ode2check.solution)
            #ode2check._as_fode()
            #ode2check.solution
            C_const_sol = ode2check.solution.as_explicit_dict()
            

            CodeFlowLogger(C_const_sol,'tut C_const_sol',self)
            
        else:
            C_const_sol={}
        # print('tut')
        for order, approx_sol in enumerate(gen_sol):
            
            # print('stale: ')
            # display(C_const_sol)
            # print('aproox bez subsa')
            
            
            # display(approx_sol.rhs.applyfunc(lambda obj: obj.expand().doit().subs(C_const_sol).subs(C_const_sol)
            #                                ))
            solution = approx_sol.applyfunc(lambda obj: obj.expand().doit()).subs(
                {
                    #self.t_list[1]: self.eps  * self.ivar
                 })

            # print('aproox ze subsa')
            # display(solution)

            sol_list += [(self.eps**order) *
                            solution.rhs.subs({
                                # self.t_list[1]: self.eps  * self.ivar,
                                # self.t_list[0]: self.ivar
                            })   ]
            
        #display(*list(SimplifiedExpr._subs_container.values()))
        result = (sum(sol_list, Matrix(2*len(self.dvars)*[0])  )).applyfunc(lambda obj: obj.expand().doit())
        
        new_res = PerturbationODESolution.from_vars_and_rhs(Matrix(list(self.dvars) +  list(self.dvars.diff(self.ivar)) ) , result)#.set_small_parameter(self.eps)
        new_res.small_parameter = self.eps
        new_res.ivar = self.ivar

        return new_res

    @property
    def _consts_dict_in_time(self):

        
        order = self.order
        
        #print('my order',order)
        sol_list = []
        gen_sol = self._general_sol(order)
        

        
        if self.eps in self.secular_eq:

            ode2check = self.secular_eq[self.eps]#.as_first_ode_linear_system()

            

            C_const_sol = ode2check.solution.as_explicit_dict()
            

        else:
            C_const_sol={}
        # print('tut')

            
        # #display(*list(SimplifiedExpr._subs_container.values()))
        # result = (sum(sol_list, Matrix(2*len(self.dvars)*[0])  )).applyfunc(lambda obj: obj.expand().doit())
        
        # new_res = PerturbationODESolution.from_vars_and_rhs(Matrix(list(self.dvars) +  list(self.dvars.diff(self.ivar)) ) , result)#.set_small_parameter(self.eps)
        # new_res.small_parameter = self.eps
        # new_res.ivar = self.ivar

        scales_dict = {
                                self.t_list[1]: self.eps  * self.ivar,
                                self.t_list[0]: self.ivar
                            }

        C_const_sol = {key:const_func.subs(scales_dict)   for key,const_func in C_const_sol.items()}

        new_res = AnalyticalSolution.from_vars_and_rhs(list(C_const_sol.keys()),list(C_const_sol.values()))#.set_small_parameter(self.eps)
        new_res.ivar = self.ivar


        return new_res


    @property
    def general_solution(self):

        
        order = self.order
        
        #print('my order',order)
        sol_list = []
        gen_sol = self._general_sol(order)
        
        # print('gen_sol')
        # display(*gen_sol)
        
        if self.eps in self.secular_eq:

            ode2check = self.secular_eq[self.eps]#.as_first_ode_linear_system()

            
            CodeFlowLogger(ode2check,'sec_eq',self)
            
            # display(ode2check._as_fode())
            # display(ode2check.solution)
            #ode2check._as_fode()
            #ode2check.solution
            C_const_sol = ode2check.solution.as_explicit_dict()
            

            CodeFlowLogger(C_const_sol,'tut C_const_sol',self)
            
        else:
            C_const_sol={}
        # print('tut')
        for order, approx_sol in enumerate(gen_sol):
            
            # print('stale: ')
            # display(C_const_sol)
            # print('aproox bez subsa')
            
            
            # display(approx_sol.rhs.applyfunc(lambda obj: obj.expand().doit().subs(C_const_sol).subs(C_const_sol)
            #                                ))
            solution = approx_sol.applyfunc(lambda obj: obj.expand().doit().subs(C_const_sol).subs(C_const_sol)).subs(
                {self.t_list[1]: self.eps  * self.ivar})

            # print('aproox ze subsa')
            # display(solution)

            sol_list += [(self.eps**order) *
                            solution.rhs.subs({
                                self.t_list[1]: self.eps  * self.ivar,
                                self.t_list[0]: self.ivar
                            })   ]
            
        #display(*list(SimplifiedExpr._subs_container.values()))
        result = (sum(sol_list, Matrix(2*len(self.dvars)*[0])  )).applyfunc(lambda obj: obj.expand().doit())
        
        new_res = PerturbationODESolution.from_vars_and_rhs(Matrix(list(self.dvars) +  list(self.dvars.diff(self.ivar)) ) , result)#.set_small_parameter(self.eps)
        new_res.small_parameter = self.eps
        new_res.ivar = self.ivar

        return new_res

    @cached_property
    def steady_solution(self):

        result = Matrix(2*len(self.dvars)*[0])

        new_res = PerturbationODESolution(Matrix(list(self.dvars) +  list(self.dvars.diff(self.ivar)) ) , result)#.set_small_parameter(self.eps)
        new_res.small_parameter = self.eps
        new_res.ivar = self.ivar
        return new_res

    @cached_property
    def solution(self):

        return self.general_solution


    def compute_solution(   
                        self,
                        t_span=None,
                        ic_list=None,
                        t_eval=None,
                        params_values=None,
                        method='RK45'
                        ):

        #         print('compute_solution')
        #         display(params_values)
        #         display(ic_list)

        if ic_list:
            print('class ics has been taken')
            self.ics = ic_list

        return self.__call__(ivar=t_span,
                             order=self._order,
                             params_values=params_values)


    def multiple_eps_exc(self, eps_array, exc_dict, fixed_params, ics, t_span):
        """
        Perform simulations for different \epsilon values and generate DataFrame comparisons 
        between ODESystem and MultiTimeScaleSolution results.

        Parameters:
            eps_array (list): List of \epsilon values for the simulations.
            exc_dict (dict): Dictionary to substitute natural frequency.
            fixed_params (dict): Dictionary of other fixed parameters (e.g., c, k, m).
            ics (dict): Initial conditions for the system.
            t_span (array-like): Time span for the simulations.

        Returns:
            comparisons (list): List of TimeDataFrame comparisons for each \epsilon.
        """
        # Define symbols
        c = Symbol('c', positive=True)
        k = Symbol('k', positive=True)
        m = Symbol('m', positive=True)
        eps = Symbol('\epsilon')
        omega = Symbol('\omega', positive=True)
        omega_0 = Symbol('omega_0', positive=True)
        Omega = Symbol('Omega', positive=True)
        delta = Symbol('delta', positive=True)

        # Initialize systems with substitutions
        msm_exc = MultiTimeScaleSolution(self.odes.subs(exc_dict), self.dvars, eps=self.eps)
        ode_exc = ODESystem(self.odes.subs(exc_dict), self.dvars)

        msm_sim = []  # List for MSM simulations
        ode_sim = []  # List for ODE simulations

        # Construct `data_arr` using eps_array and fixed_params
        data_arr = [{**fixed_params, eps: epsilon} for epsilon in eps_array]

        # Loop through the data array and compute solutions
        for value in data_arr:
            msm_sim_i = msm_exc.solution.subs(value).expand().with_ics(ics).expand().n(3).numerized().compute_solution(t_span)
            ode_sim_i = ode_exc.solution.subs(value).expand().with_ics(ics).expand().n(3).numerized().compute_solution(t_span)

            # Append to the respective arrays
            msm_sim.append(msm_sim_i)
            ode_sim.append(ode_sim_i)

        # Dynamically create comparisons for all eps values in data_arr
        comparisons = []
        for i, epsilon in enumerate(eps_array):
            comparison = TimeDataFrame({
                (f'ODESystem', f'\epsilon={epsilon}', self.dvars[0]): ode_sim[i].iloc[:, 0],
                (f'MultiScaleSolution',  f'\epsilon={epsilon}', self.dvars[0]): msm_sim[i].iloc[:, 0],
            })
            comparisons.append(comparison)

        return comparisons

    def multiple_eps_unexcited(self, eps_array, exc_dict, fixed_params, ics, t_span):
        """
        Perform simulations for different \epsilon values and generate DataFrame comparisons 
        between ODESystem and MultiTimeScaleSolution results.

        Parameters:
            eps_array (list): List of \epsilon values for the simulations.
            exc_dict (dict): Dictionary to substitute natural frequency.
            fixed_params (dict): Dictionary of other fixed parameters (e.g., c, k, m).
            ics (dict): Initial conditions for the system.
            t_span (array-like): Time span for the simulations.

        Returns:
            comparisons (list): List of TimeDataFrame comparisons for each \epsilon.
        """
        # Define symbols
        c = Symbol('c', positive=True)
        k = Symbol('k', positive=True)
        m = Symbol('m', positive=True)
        eps = Symbol('\epsilon')
        omega = Symbol('\omega', positive=True)
        omega_0 = Symbol('omega_0', positive=True)
        Omega = Symbol('Omega', positive=True)
        delta = Symbol('delta', positive=True)

        # Initialize systems with substitutions
        msm_exc = self
        ode_exc = ODESystem(self.odes.subs(exc_dict), self.dvars)

        msm_sim = []  # List for MSM simulations
        ode_sim = []  # List for ODE simulations

        # Construct `data_arr` using eps_array and fixed_params
        data_arr = [{**fixed_params, eps: epsilon} for epsilon in eps_array]

        # Loop through the data array and compute solutions
        for value in data_arr:
            msm_sim_i = msm_exc.solution.subs(value).expand().with_ics(ics).expand().n(3).numerized().compute_solution(t_span)
            ode_sim_i = ode_exc.solution.subs(value).expand().with_ics(ics).expand().n(3).numerized().compute_solution(t_span)

            # Append to the respective arrays
            msm_sim.append(msm_sim_i)
            ode_sim.append(ode_sim_i)

        # Dynamically create comparisons for all eps values in data_arr
        comparisons = []
        for i, epsilon in enumerate(eps_array):
            comparison = TimeDataFrame({
                (f'Using ODESystem, {self.dvars[0]}', f'\epsilon={epsilon}'): ode_sim[i].iloc[:, 0],
                (f'Using MultiScaleSolution, {self.dvars[0]}', f'\epsilon={epsilon}'): msm_sim[i].iloc[:, 0],
            })
            comparisons.append(comparison)

        return comparisons
    
    
    def _numerical_comparison(self,t_span, eps_array=[0.1], fixed_params=None, ics=None):
        """
        Perform simulations for different \epsilon values and generate DataFrame comparisons 
        between ODESystem and MultiTimeScaleSolution results.

        Parameters:
            eps_array (list): List of \epsilon values for the simulations.
            exc_dict (dict): Dictionary to substitute natural frequency.
            fixed_params (dict): Dictionary of other fixed parameters (e.g., c, k, m).
            ics (dict): Initial conditions for the system.
            t_span (array-like): Time span for the simulations.

        Returns:
            comparisons (list): List of TimeDataFrame comparisons for each \epsilon.
        """
        # Define symbols
        c = Symbol('c', positive=True)
        k = Symbol('k', positive=True)
        m = Symbol('m', positive=True)
        eps = self.eps
        omega = Symbol('\omega', positive=True)
        omega_0 = Symbol('omega_0', positive=True)
        Omega = Symbol('Omega', positive=True)
        delta = Symbol('delta', positive=True)

        if fixed_params is None:
            fixed_params = {}

        # Initialize systems with substitutions
        msm_exc = self
        ode_exc = ODESystem(self.odes, self.dvars)

        msm_sim = []  # List for MSM simulations
        ode_sim = []  # List for ODE simulations


        if ics is None:
            ics = [0.0,0.0]
        
        # Construct `data_arr` using eps_array and fixed_params
        data_arr = [{**fixed_params, eps: epsilon} for epsilon in eps_array]

        # Loop through the data array and compute solutions
        for value in data_arr:
            msm_sim_i = msm_exc.solution.subs(value).expand().with_ics(ics).expand().n(6).numerized().compute_solution(t_span)
            ode_sim_i = ode_exc.subs(value).n(6).numerized().with_ics(ics).compute_solution(t_span)

            # Append to the respective arrays
            msm_sim.append(msm_sim_i)
            ode_sim.append(ode_sim_i)

        # Dynamically create comparisons for all eps values in data_arr
        comparisons = []
        for i, epsilon in enumerate(eps_array):
            comparison = TimeDataFrame({
                (f'Using ODESystem, {self.dvars[0]}', f'\epsilon={epsilon}'): ode_sim[i].iloc[:, 0],
                (f'Using MultiScaleSolution, {self.dvars[0]}', f'\epsilon={epsilon}'): msm_sim[i].iloc[:, 0],
            })
            comparisons.append(comparison)

        return comparisons